# ------------------------------------------------------------
# Auto generated test file `test_060_SimulationParameters.py`
# ------------------------------------------------------------
# source file   = NBTest_060_SimulationParameters.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 060
# test comment  = SimulationParameters
# ------------------------------------------------------------



from carbon.helpers.stdimports import *
from carbon.routers import ExactBase, ExactRouterX0Y0N, AlphaRouter, FastRouter
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonSimulatorUI))
print_version(require="2.3.3")




# ------------------------------------------------------------
# Test      060
# File      test_060_SimulationParameters.py
# Segment   New parameter handling
# ------------------------------------------------------------
def test_new_parameter_handling():
# ------------------------------------------------------------
    #
    # In v2.3.3 we introduced a `parameters` dict into `BaseRouter` that is collecting all simulation parameters, instead of adding them individually as function parameters. The parameters removed are the following:
    #
    # - `debug`
    # - `verbose`
    # - `raiseonerror`
    # - `assert_precision`
    # - `use_floor_division`
    #
    #
    # Note that for compatibility reasons all those paramters can still be accessed as properties, as well through the regular `p("...")` pathway.
    
    ROUTERS = (ExactBase, ExactRouterX0Y0N, AlphaRouter, FastRouter)
    #ROUTERS = (ExactBase, ExactRouterX0Y0N, AlphaRouter)
    
    for Router in ROUTERS:
        print(Router)
        router = Router()
        assert router.parameters == {}
        assert router.debug == False
        assert router.verbose == False
        assert router.raiseonerror == False
        assert router.assert_precision == 4
        assert router.use_floor_division == False
    
        params = {
            "debug": True,
            "verbose": True,
            "raiseonerror": True,
            "assert_precision": 5,
            "use_floor_division": True,
        }
        router = Router(parameters=params)
        assert router.parameters == params
        assert router.debug == True
        assert router.verbose == True
        assert router.raiseonerror == True
        assert router.assert_precision == 5
        assert router.use_floor_division == True

# ------------------------------------------------------------
# Test      060
# File      test_060_SimulationParameters.py
# Segment   CarbonSimulatorUI
# ------------------------------------------------------------
def test_carbonsimulatorui():
# ------------------------------------------------------------
    #
    # Make sure the Simulation object is working correctly
    
    Sim = CarbonSimulatorUI()
    assert Sim.matcher.parameters == {
        'debug': False,
        'verbose': False,
        'raiseonerror': False,
        'assert_precision': 4,
        'use_floor_division': False
    }
    
    Sim = CarbonSimulatorUI(debug=True)
    assert Sim.matcher.p("debug") == True
    
    Sim = CarbonSimulatorUI(verbose=True)
    assert Sim.matcher.p("verbose") == True
    
    Sim = CarbonSimulatorUI(raiseonerror=True)
    assert Sim.matcher.p("raiseonerror") == True
    
    pass
    
    
    
    
    
    


"""
Carbon helper module -- encapsulate parameters for a single strategy
"""
__VERSION__ = "2.7"
__DATE__ = "02/02/2023"


from dataclasses import dataclass as _dataclass
from math import sqrt as _sqrt
from copy import copy as _copy
from math import isnan as _isnan

@_dataclass
class strategy():
    """
    class describing a Carbon strategy

    :p_sell_a:          price a (=start) of the sell/ask range*  
    :p_sell_b:          price b (=end) of the sell/ask range*  
    :p_buy_a:           price a (=start) of the buy/bid range*   
    :p_buy_b:           price b (=end) of the buy/bid range*   
    :amt_rsk:           the amount of risk asset the strategy is seeded with**
    :amt_csh:           the amount of cash the strategy is seeded with**
    :psell_marginal:    the current marginal price of the sell range
    :pbuy_marginal:     the current marginal price of the buy range
    :rsk:               risk asset name (default RSK)
    :csh:               cash name (default CSH)
    :rescale:           whether or not the strategy should be rescaled when
                        calling the rescale method
    :rescalted_fctr:    the factor this strategy was rescaled (None if not rescaled)

    *p_bid_x and p_ask_x are provided as alias properties

    **note that strategies that are seeded with 0 are instead seeded with
    `cls.MIN_SEED_AMT` which is set to 1e-10 (it can be changed in a derived
    class). The reason is that zero-seeded ranges do not allow for marginal prices.
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    p_sell_a: float
    p_sell_b: float 
    p_buy_a: float
    p_buy_b: float
    amt_rsk: float = None
    amt_csh: float = None

    psell_marginal:float = None
    pbuy_marginal:float = None

    y_int_sell:float = None
    y_int_buy:float = None
        
    rsk: str = "RSK"
    csh: str = "CSH"

    rescale: bool = True
    rescaled_fctr: float = None

    def __post_init__(self):
        pm_provided = not self.psell_marginal is None or not self.pbuy_marginal is None
        yi_provided = not self.y_int_sell is None or not self.y_int_buy is None
        if pm_provided and yi_provided:
            raise ValueError("Not both p_marginal and y_int can be provided", self.psell_marginal, self.y_int_sell, self.pbuy_marginal, self.y_int_buy)
        
    MIN_SEED_AMT = 1e-10

    def rescale_strat(self, newspot=100, oldspot=100, force=False):
        """
        rescale the strategy to another spot price (returns new object)

        :newspot:     the new spot price  to which to rescale
        :oldspot:     the old spot price basis for rescaling
        :force:       normally, strategies with rescale = False are not rescaled
                      using force = True forces a rescale
        """
        if not self.rescale:
            if not force:
                #print("[rescale_strat] not rescaling", newspot, oldspot, self)
                return _copy(self)
            else:
                print("[rescale_strat] forcing rescale", newspot, oldspot, self)

        fctr = newspot / oldspot
        #print(f"[rescale_strat] rescaling new={newspot}, old={oldspot}, fctr={fctr}", self)
        return self.__class__(
            p_sell_a = self.p_sell_a*fctr,
            p_sell_b = self.p_sell_b*fctr,
            p_buy_a  = self.p_buy_a*fctr,
            p_buy_b  = self.p_buy_b*fctr,

            amt_rsk  = self.amt_rsk,
            amt_csh  = self.amt_csh,

            psell_marginal  = self.psell_marginal*fctr if not self.psell_marginal is None else None,
            pbuy_marginal   = self.pbuy_marginal*fctr if not self.pbuy_marginal is None else None,

            # it is not really clear what to do here; strategies
            # that do set yint should probably just set "do not rescale"
            # we may come up with a better solution once we have marginal
            # prices sorted in strategy creation
            y_int_sell  = self.y_int_sell,  
            y_int_buy = self.y_int_buy*fctr if not self.y_int_buy is None else None,   # CSH number

            rsk  = self.rsk,
            csh  = self.csh,

            rescale = False,
            rescaled_fctr = fctr,
        )

    def set_tvl(self, spot, cashpc, tvl):
        """
        creates new strategy and sets amt_csh and amt_rsk based on TVL, pc cash and spot

        :spot:      the spot price (in CSH per RSK)
        :cashpc:    the percentage of the portfolio in CSH (1.0=100%)
        :tvl:       the total portfolio value in CSH (default=1,000)
        :returns:   in every case this function returns a new object;
                    provided both amt_rsk and amt_csh were None (not 0!)
                    they are changed; otherwise newobj = self
        """
        #print("[set_tvl] spot, cashpc, tvl" , spot, cashpc, tvl)
        newobj = _copy(self)
        if self.amt_rsk is None and self.amt_csh is None:
            #print("[set_tvl] setting new values for amt_rsk, amt_csh")
            newobj.amt_rsk = tvl/spot*(1-cashpc)
            newobj.amt_csh = tvl*cashpc
        else:
            #print("[set_tvl] existing values for amt_rsk, amt_csh; won't touch")
            pass
        return newobj

    def __post_init__(self):
        # the ==0 is important: do not touch None!
        # (None means: to be filled in later)
        if self.amt_rsk == 0: self.amt_risk = self.MIN_SEED_AMT
        if self.amt_csh == 0: self.amt_csh  = self.MIN_SEED_AMT
        if self.rsk is None: self.rsk = "RSK"
        if self.csh is None: self.csh = "CSH"

    @classmethod
    def from_dct(cls, dct, raiseonerror=True):
        """
        alternative constructor: reconstruct class from dict

        :dct:               a big dict, typically read from a full scenario description
        :raiseonerror:      if True, raises on error, otherwise returns None

        NOTE: the dict fields expected all start with `strat_`, and `strat_type`
        indicates which constructor to call; eg `mgw` expects data records suitable
        for the mgw constructor. For exact field names expected, please check
        the source code of this method
        """
        r = dct
        try:
            if r["strat_type"].lower() == "mgw":
                try:
                    w = (r["strat_wsell"], r["strat_wbuy"])
                except:
                    w = r["strat_w"]
                try:
                    u = (r["strat_usell"], r["strat_ubuy"])
                except:
                    u = r["strat_u"]
                return cls.from_mgw(m=r["strat_m"], g=r["strat_g"], w=w, u=u)

            elif r["strat_type"].lower() == "fixed":
                return cls(
                    r["p_sell_a"], 
                    r["p_sell_b"], 
                    r["p_buy_a"], 
                    r["p_buy_b"], 
                    rescale=False,
                )

            elif r["strat_type"].lower() == "uv3": 
                return cls.from_u3(
                    p_lo=r["p_lo"],
                    p_hi=r["p_hi"],
                    start_below=bool(r["start_below"]),
                    fee_pc=r["fee_pc"],
                )
                
            else:
                raise ValueError(f"Unknown strategy {dct['strat_type']}")
        except:
            if raiseonerror:
                raise
            return None
        
    MAX_UTIL = 0.999 # all MAXUTIL < u <= 1 are set to u = MAXUTIL
    
    @classmethod
    def _validate(cls, u):
        """validates the u value for from_mgw"""
        if not u: return None
        if u<0:
            raise ValueError("Must have u >= 0", u)
        elif u>1:
            raise ValueError("Must have u < 1", u)
        elif u > cls.MAX_UTIL:
            u = cls.MAX_UTIL
        return u

    @staticmethod
    def _sorted(x1, x2, inverse=False):
        """
        returns x1,x2 in order x1<x2 (x1>x2 if inverse)
        """
        invert = (x2 > x1) ^ bool(inverse) 
        return (x1,x2) if invert else (x2,x1)
    
    @classmethod 
    def from_sb(cls, sell, buy, fixed=True, overlapping=False):
        """
        alternative constructor: create instance from sell range and buy range

        :sell:          the sell (ask) range, as tuple (start=lo, end=hi)
        :end:           the buy (bid) range, as tuple (start=him end=lo)
        :fixed:         whether the range is fixed or rescaleable
        :overlapping:   if False (default), raises on overlapping ranges
        """
        try:
            p_sell_a, p_sell_b = cls._sorted(*sell)
        except TypeError:
            p_sell_a = p_sell_b = sell

        try:
            p_buy_a, p_buy_b = cls._sorted(*buy, inverse=True)
        except TypeError:
            p_buy_a = p_buy_b = buy

        if p_buy_a > p_sell_a and not overlapping:
            raise ValueError("Must not have overlapping ranges unless overlapping True", sell, buy, p_sell_a, p_buy_a)

        return cls(
            p_sell_a = p_sell_a,
            p_sell_b = p_sell_b,
            p_buy_a = p_buy_a,
            p_buy_b = p_buy_b,  
            rescale = not fixed,    
         )

    @classmethod
    def from_mgw(cls, m=100, g=0, w=0, u=0, amt_rsk=None, amt_csh=None, rsk=None, csh=None):
        """
        alternative constructor: create instance from mid `m`, gap width `g`, and range width `w`
        
        :m:             (geometric) middle between the two ranges
        :g:             geometric gap between the two ranges
        :w:             geometric width of the range
        :u:             percentage utilization of the range (0.01=1%; 0<=u<1)**
                        note: a tuple/list is also accepted, in which case it
                        is (usell, ubuy)
        :amt_rsk:       amount of risk asset the strategy is seeded with*
        :amt_csh:       amount of cash the strategy is seeded with*
        :rsk:           risk asset name (default RSK)
        :csh:           cash name (default CSH)

        *note that strategies that are seeded with 0 are instead seeded with
        `cls.MIN_SEED_AMT` which is set to 1e-10 (it can be changed in a derived
        class). The reason is that zero-seeded ranges do not allow for marginal prices.

            p_buy_a     = m/(1+g/2)
            p_buy_b     = m/(1+g/2)/(1+w)
            p_sell_a    = m*(1+g/2)
            p_sell_b    = m*(1+g/2)*(1+w)

        **a standard range has u=0, meaning that p_marg = p_start; when u->1 we have 
        p_marg -> p_end (we can not have u=1 for y>0); p_marg = (1-u) p_s + u p_e;
        for values MAX_UTIL < u <=1, u is set to MAX_UTIL
        """
        if w==0: w=0.0001
        if g==0: g=0.0001
        try:
            # this will work on tuple / list of length 2...
            usell, ubuy = (cls._validate(uu) for uu in u)
        except:
            # ...and this for a single one
            usell = ubuy = cls._validate(u)
        #print("[from_mgw] u, usell, ubuy", u, usell, ubuy)
    
        try:
            wsell, wbuy = (cls._validate(ww) for ww in w)
        except:
            wsell = wbuy = cls._validate(w)
        #print("[from_mgw] w, wsell, wbuy", w, wsell, wbuy)
    
        
        g2 = 0.5*g
        p_buy_a     = m/(1+g2)
        p_buy_b     = m/(1+g2)/(1+wbuy)
        p_sell_a    = m*(1+g2)
        p_sell_b    = m*(1+g2)*(1+wsell)

        return cls(
            p_buy_a = p_buy_a, 
            p_buy_b = p_buy_b,
            p_sell_a = p_sell_a,
            p_sell_b = p_sell_b,
            amt_rsk = amt_rsk,
            amt_csh = amt_csh,
            rsk = rsk,
            csh = csh,
            psell_marginal = (1-usell)*p_sell_a + usell*p_sell_b if usell else None,
            pbuy_marginal = (1-ubuy)*p_buy_a + ubuy*p_buy_b if ubuy else None,
        )
    
    @classmethod
    def from_u3(cls, p_lo, p_hi, start_below, tvl_csh=None, fee_pc=None, rsk=None, csh=None):
        """
        creates a strategy equivalent to a uniswap v3 position

        :p_lo:          the low price of the range
        :p_hi:          the high price of the range
        :fee_pc:        percentage fees (0.1=1%)
        :start_below:   if True/False, the start price is below/above the range, so 
                        the combined range is 100% in RSK/CSH p_marginal at p_lo/p_hi
        :tvl_csh:       the TVL (rsk+csh) at top of the range, measured in csh (default: 1000)
        :rsk:           risk asset name (default RSK)
        :csh:           cash name (default CSH)
        """
        #assert fee_pc is None, "fees not implemented yet"
        if not p_hi >= p_lo:
            raise ValueError("Must have p_hi > p_lo", p_hi, p_lo)
        if tvl_csh is None:
            tvl_csh = 1000

        # the SELL range is always the (red) UPPER range
        # for the sell range pa < pb
        # the BUY range is always the (green) LOWER range
        # for the buy range pa > pb
        p_sell_a = p_buy_b = p_lo   
        p_sell_b = p_buy_a = p_hi

        if not fee_pc is None:
            #fee_mult = 1+0.5*fee_pc
            fee_shift = _sqrt(p_lo*p_hi)*fee_pc*0.5
            # p_sell_a *= fee_mult
            # p_sell_b *= fee_mult
            # p_buy_a  /= fee_mult
            # p_buy_b  /= fee_mult
            p_sell_a += fee_shift
            p_sell_b += fee_shift
            p_buy_a  -= fee_shift
            p_buy_b  -= fee_shift

        if start_below:
            #p_marginal = p_lo*1.0000000001
            amt_rsk = tvl_csh/_sqrt(p_lo*p_hi)
            amt_csh = 0
        else:
            #p_marginal = p_hi*0.9999999999
            amt_rsk = 0
            amt_csh = tvl_csh

        return cls(
            p_buy_a = p_buy_a, 
            p_buy_b = p_buy_b,
            p_sell_a = p_sell_a,
            p_sell_b = p_sell_b,
            amt_rsk = amt_rsk,
            amt_csh = amt_csh,
            rsk = rsk,
            csh = csh,
            # psell_marginal = p_marginal*fee_mult,
            # pbuy_marginal = p_marginal/fee_mult,
            y_int_sell = tvl_csh/_sqrt(p_lo*p_hi), # RSK NUMBER
            y_int_buy  = tvl_csh, # CSH NUMBER

            # rescale = False is extremely important as rescaling
            # typically changes the finely balanced ratios
            rescale = False,
        )

    @property
    def descr(s):
        """returns a description of the strategy"""
        bid_s = f"BID {s.p_buy_b:.1f}-{s.p_buy_a:.1f} [{s.amt_csh} {s.csh}]"
        ask_s = f"ASK {s.p_sell_a:.1f}-{s.p_sell_b:.1f} [{s.amt_rsk} {s.rsk}]"
        mid_s = f"MID {_sqrt(s.p_buy_a*s.p_sell_a):.1f}"
        return f"{bid_s} - {mid_s} - {ask_s}"
    
    @property
    def p(s):
        """DEPRECTATED returns a tuple suitable for `add_strategy (*params)`"""
        print("[strategy::p] DEPRECATED. PLEASE USE dct INSTEAD")
        return (s.rsk, s.amt_rsk, s.p_sell_a, s.p_sell_b, s.amt_csh, s.p_buy_a, s.p_buy_b)
    
    @property
    def dct(s):
        """returns a dict suitable for `add_strategy (**params)`"""
        return {
            "tkn":              s.rsk, 
            "amt_sell":         s.amt_rsk, 
            "psell_start":      s.p_sell_a, 
            "psell_end":        s.p_sell_b, 
            "amt_buy":          s.amt_csh, 
            "pbuy_start":       s.p_buy_a, 
            "pbuy_end":         s.p_buy_b,
            "pair":             s.slashpair,
            "psell_marginal":   s.psell_marginal,
            "pbuy_marginal":    s.pbuy_marginal,
            "y_int_sell":       s.y_int_sell,
            "y_int_buy":        s.y_int_buy,
        }
    
    @property
    def slashpair(self):
        """returns the slashpair as str, eg 'RSK/CSH'"""
        return f"{self.rsk}/{self.csh}"

    @staticmethod
    def rr(df):
        """
        returns the data frame df as dict {index -> {field -> value}}

        :df:    in this context, the data frame holding all key parameters of a chart 
        """
        rr = df.to_dict(orient="index")
        return rr

    @staticmethod
    def r0(rr):
        """
        returns the first element of the dict `rr` (here: the reference record)
        """
        return rr[tuple(rr)[0]]

    @staticmethod
    def dr(r, r0, purge=True):
        """
        difference between the record `r` and baseline record `r0`

        :r:         the record to be evaluated
        :r0:        baseline record
        :purge:     if False, fields that don't change contain None; otherwise purged
        :returns:   the difference record as dict
        """
        result = {k: r[k] if r[k] != r0[k] else None for k in r0}
        if not purge: return result
        return {k:v for k,v in result.items() if not v is None and not k in ["id0", "comment"]}

    @staticmethod
    def drr(rr, baseline=False, purge=True, asdf=False, comment=False):
        """
        run `dr` on dict {id -> r[id]}

        :rr:            the dict of records dict {id -> r[id]}
        :baseline:      if True, add content for baseline (=first) record r0
        :purge:         same as in `dr`
        :asdf:          if True, result returned as data frame, otherwise dict
        :comment:       if False does never return comment column in the dataframe
        :returns:       dict or data frame, depending on asdf
        """
        r0k = tuple(rr)[0]
        r0 = rr[r0k]
        if not asdf:
            result = {k:dr(r, r0, purge) for k,r in rr.items()}
            if not baseline:
                return result
            result[r0k] = {k:v for k,v in r0.items() if not k in ["id0", "comment"]}
            return result
        else:
            result = {k:dr(r, r0, purge=False) for k,r in rr.items()}
            df = pd.DataFrame.from_records(result).T.fillna("")
            del df["id0"]
            if not comment: 
                del df["comment"]
            df = df[ [col for col in df.columns  if (df[col]!= "").any()] ]                
            return df

    @staticmethod
    def filter_isnan(val):
        """
        returns val unless val==nan, then it returns ''
        """
        try:
            if _isnan(val):
                return ""
        except TypeError:
            return val


from .sdktoken import SDKToken, Tokens
from ..pair import CarbonPair
from .carbonsdk0 import CarbonSDK0
from dataclasses import dataclass
from math import sqrt


#########################################################################################################
##  DATA CLASSES
@dataclass
class PairLiquidity:
    """
    a range object for a specific pair

    :amount:    the amount of liquidity in the pair (float or tuple of floats)
    :unit:      the unit of the liquidity amount (as token name)
    :AMMsells:  the token the AMM sells (as token name)
    :bidAsk:    the bid/ask direction (as string)
    :pair:      the pair (as CarbonPair object)
    :rate:      the rate of the quoted liquidity (None = total; float or tuple of floats)
    :inverted:  whether the rate is inverted compared to the raw API values
    :islist:    whether the liquidity is a list (tuple of floats)
    """
    amount: any
    unit: str
    AMMsells: str
    bidAsk: str
    pair: CarbonPair
    rate: any=None
    inverted: bool=None
    islist: bool=None

@dataclass
class RangeByPair:
    """
    a range object for a specific pair

    :startRate:     the exchange rate at which selling starts, in units of the pair
    :endRate:       the exchange rate at which selling ends, in units of the pair
    :AMMsells:      the token the AMM sells (as token name)
    :AMMbuys:       the token the AMM buys (as token name)
    :bidAsk:        the bid/ask direction (as string)
    :pair:          the pair (as CarbonPair object)
    :inverted:      whether the price is inverted as compared to the raw API calls
    """
    startRate: float
    endRate: float
    AMMsells: str
    AMMbuys: str
    bidAsk: str
    pair: CarbonPair
    inverted: bool

@dataclass
class MarketByPair():
    """
    market overview for a given pair

    :pair:              pair (as str)
    :bestBid:           best bid available in the market
    :bestAsk:           best ask available in the market
    :liqBid:            liquidity on the bid side of the market
    :liqAsk:            liquidity on the ask side of the market 
    :lastBid:           last bid available in the market
    :lastAsk:           last ask available in the market
    ----
    :inverted:          True if the pair is inverted (bid > ask)
    :mid:               mid price
    :spread:            spread in price units
    :spreadpc:          spread as a percentage of the mid price
    :cpair:             associated CarbonPair object
    :price_convention:  price convention of all prices in this object
    :liqBidUnit:        unit of the bid liquidity (=tknq)
    :liqAskUnit:        unit of the ask liquidity (=tknb)
    :bidAMMSells:       token sold by AMM on the bid side (=tknq)
    :askAMMSells:       token sold by AMM on the ask side (=tknb)
    :bidTraderSells:    token sold by Trader on the bid side (=tknb)
    :askTraderSells:    token sold by Trader on the ask side (=tknq)
    """
    pair: str
    bestBid: float
    bestAsk: float
    liqBid: float
    liqAsk: float
    lastBid: float
    lastAsk: float
    
    @property
    def inverted(self):
        return self.bestBid > self.bestAsk
    
    @property
    def mid(self):
        return 0.5*(self.bestBid+self.bestAsk)
    
    @property
    def spreadpc(self):
        return self.spread/self.mid
    
    @property
    def spread(self):
        return self.bestAsk-self.bestBid
    
    @property
    def cpair(self):
        return CarbonPair(self.pair)
    
    @property
    def price_convention(self):
        return self.cpair.price_convention
    
    @property
    def liqBidUnit(self):
        return self.cpair.tknq
    
    @property
    def liqAskUnit(self):
        return self.cpair.tknb
    
    @property
    def liqBidUnit(self):
        return self.cpair.tknq
    
    @property
    def bidAMMSells(self):
        return self.cpair.tknq
    
    @property
    def askAMMSells(self):
        return self.cpair.tknb
    
    @property
    def bidTraderSells(self):
        return self.cpair.tknb
    
    @property
    def askTraderSells(self):
        return self.cpair.tknq
    
    def ibid(self, x):
        """
        interpolate bid price (x=0: best bid; x=1: last bid)
        """
        return self.bestBid + x*(self.lastBid-self.bestBid)
    
    def iask(self, x):
        """
        interpolate ask price (x=0: best ask; x=1: last ask)
        """
        return self.bestAsk + x*(self.lastAsk-self.bestAsk)
    
    def iall(self, x):
        """
        interpolate all prices (x=0: last bid; x=1: last ask)
        """
        return self.lastBid + x*(self.lastAsk-self.lastBid)
         
@dataclass
class TradeData:
    """
    an object containing the data needed to execute a trade

    :pair:                  the pair to which the trade applies
    :tkn:                   the token specified by the trader
    :traderBuySell:         whether this token is bought or sold by the trader
    :amount:                the amount of tkn to be bought or sold
    :sourceToken:           the token that the trader sells (as SDKToken object)
    :targetToken:           the token that the trader buys (as SDKToken object)
    :tradeActions:          the trade actions (input for the next step)
    :actionsWei:            the trade actions in wei
    :actionsTokenRes:       the trade actions in token resolution
    :tradeByTargetAmount:   corresponding SDK parameter
    """
    pair: str
    tkn: str
    traderBuySell: str
    amount: float
    sourceToken: SDKToken 
    targetToken: SDKToken
    tradeByTargetAmount: bool
    tradeActions: list
    actionsWei: list
    actionsTokenRes: list
    tradeData: dict

    @property
    def cttkwargs(self):
        """kwargs for mComposeTradeTransaction"""
        return dict(
            sourceToken=self.sourceToken,
            targetToken=self.targetToken,
            tradeActions=self.tradeActions,
            tradeData=self.tradeData,
            tradeByTargetAmount=self.tradeByTargetAmount,
        )
    
    @property
    def cttargs(self):
        """args for mComposeTradeTransaction"""
        return tuple(self.cttkwargs.values())
    
@dataclass
class EncodedOrder():
    """
    a single curve as encoded by the SDK

    :y:          number of token wei to sell on the curve
    :z:          curve capacity in number of token wei
    :A:          curve parameter A, multiplied by 2**48
    :B:          curve parameter B, multiplied by 2**48
    ----
    :A_:         curve parameter A in proper units
    :B_:         curve parameter B in proper units
    :p_start:    start token wei price of the order (in dy/dx)
    :p_end:      end token wei price of the order (in dy/dx)
    """
    token: str
    y: int
    z: int
    A: int
    B: int

    ONE = 2**48

    @classmethod
    def from_sdk(cls, token, order):
        return cls(token=token, **order)
    
    @property
    def descr(self):
        s=self
        return f"selling {s.token} @ ({1/s.p_start}..{1/s.p_end})  [TKNwei] per {s.token}wei"
        
    def __getitem__(self, item):
        return getattr(self, item)
    
    @classmethod
    def decodeFloat(cls, value):
        """undoes the mantisse/exponent encoding in A,B"""
        return (value % cls.ONE) << (value // cls.ONE)
    
    @classmethod
    def decode(cls, value):
        """decodes A,B to float"""
        return cls.decodeFloat(int(value))/cls.ONE
    
    @staticmethod
    def bitLength(value):
        "minimal number of bits needed to represent the value"
        return len(bin(value).lstrip('0b')) if value > 0 else 0
    
    @classmethod
    def encodeRate(cls, value):
        "encodes a rate float to an A,B (using cls.ONE for scaling)"
        data = int(sqrt(value) * cls.ONE)
        length = cls.bitLength(data // cls.ONE)
        return (data >> length) << length

    @classmethod
    def encodeFloat(cls, value):
        "encodes a long int value as mantisse/exponent into a shorter integer"
        exponent = cls.bitLength(value // cls.ONE)
        mantissa = value >> exponent
        return mantissa | (exponent * cls.ONE)
    
    @classmethod
    def encode(cls, y, p_start_hi, p_end_lo, p_marg=None, z=None, token=None):
        """
        alternative constructor: creates a new encoded order from the given parameters

        :y:             number of token wei currently available to sell on the curve (loading)
        :p_start_hi:    start token wei price of the order (in dy/dx; highest)
        :p_end_lo:      end token wei price of the order (in dy/dx; lowest)
        :p_marg:        marginal token wei price of the order (in dy/dx)*
        :z:             curve capacity in number of token wei*

        *at most one of p_marg and z can be given; if neither is given, 
        z is assumed to be equal to y
        """
        if token is None:
            token = "TKN"
        if not p_marg is None and not z is None:
            raise ValueError("at most one of p_marg and z can be given")
        if z is None:
            if p_marg is None:
                z = y
            else:
                if p_marg >= p_start_hi:
                    z = y
                else:
                    z = y * (p_start_hi - p_end_lo) // (p_marg - p_start_hi)
        return cls(
            y=int(y), 
            z=int(z), 
            A=cls.encodeFloat(cls.encodeRate(p_start_hi)-cls.encodeRate(p_end_lo)), 
            B=cls.encodeFloat(cls.encodeRate(p_end_lo)), 
            token=token
        )
    
    @property
    def A_(self):
        return self.decode(self.A)
    
    @property
    def B_(self):
        return self.decode(self.B)
    
    @property
    def p_end(self):
        return self.B_*self.B_
    
    @property
    def p_start(self):
        return (self.B_ + self.A_)**2
    
    @property
    def p_marg(self):
        if self.y == self.z:
            return self.p_start
        elif self.y == 0:
            return self.p_end
        raise NotImplementedError("p_marg not implemented for non-full / empty orders")
        A = self.decodeFloat(self.A)
        B = self.decodeFloat(self.B)
        return self.decode(B + A * self.y / self.z)**2
        # https://github.com/bancorprotocol/carbon-simulator/blob/beta/benchmark/core/trade/impl.py
        # 'marginalRate' : decodeRate(B + A if y == z else B + A * y / z),
        
@dataclass
class EncodedStrategy():
    """
    a single strategy as encoded by the SDK

    :sid:        strategy id
    :order0:     order selling token0 (ie y=token0, dy/dx = token0 per token1)
    :order1:     order selling token1 (ie y=token1, dy/dx = token1 per token0)
    """
    sid: int
    order0: EncodedOrder
    order1: EncodedOrder
    
    @classmethod
    def from_sdk(cls, dct):
        order0 = CarbonSDK.EncodedOrder.from_sdk(dct["token0"], dct["order0"])
        order1 = CarbonSDK.EncodedOrder.from_sdk(dct["token1"], dct["order1"])
        return cls(sid=dct["id"], order0=order0, order1=order1)

    @property
    def token0(self):
        return self.order0.token
    
    @property
    def token1(self):
        return self.order1.token
    
    @property
    def descr(self):
        de = self
        s0 = f"order0: selling {de.token0} @ ({1/de.order0.p_start}..{1/de.order0.p_end}) {de.token1}wei per {de.token0}wei"
        s1 = f"order1: buying  {de.token0} @ ({de.order1.p_start}..{de.order1.p_end}) {de.token1}wei per {de.token0}wei"
        return [s0,s1]
    
    def __getitem__(self, item):
        if item == "id": 
            return self.sid
        # elif item == "token0":
        #     return self.order0.token
        # elif item == "token1":
        #     return self.order1.token
        return getattr(self, item) 

@dataclass
class Strategy():
    """
    single strategy, as returned by the API (encoded and decoded)
    """
    sid: int
    baseToken: str
    quoteToken: str
    buyPriceLow: float
    buyPriceHigh: float
    buyBudget: float
    sellPriceLow: float
    sellPriceHigh: float
    sellBudget: float
    encoded: EncodedStrategy

    def __post_init__(self):
        assert self.baseToken == self.encoded.order0.token, "base token not equal to token0"
        assert self.quoteToken == self.encoded.order1.token, "quote token not equal to token1"

    @property
    def descr(self):
        d = self
        s0 = f"order0: selling {d.baseToken} @ ({d.sellPriceLow}..{d.sellPriceHigh}) {d.quoteToken} per {d.baseToken}"
        s1 = f"order1: buying  {d.baseToken} @ ({d.buyPriceHigh}..{d.buyPriceLow}) {d.quoteToken} per {d.baseToken}"
        return [s0,s1]
    
    @classmethod
    def from_sdk(cls, dct):
        encoded = CarbonSDK.EncodedStrategy.from_sdk(dct["encoded"])
        sid = dct["id"]
        dct1 = {k:v for k,v in dct.items() if not k in ["id", "encoded"]}
        return cls(sid=sid, encoded=encoded, **dct1)
    
    def __getitem__(self, item):
        if item == "id": 
            return self.sid 
        return getattr(self, item) 

#########################################################################################################
##  CLASS CARBON SDK
class CarbonSDK(CarbonSDK0):
    """
    additional layer of abstraction to the CarbonSDK0 class
    """

    ########################################################################
    ## data containers and constants    
    SELL = "SELL"
    BUY = "BUY"

    PairLiquidity = PairLiquidity
    RangeByPair = RangeByPair
    MarketByPair = MarketByPair
    TradeData = TradeData
    EncodedOrder = EncodedOrder
    EncodedStrategy = EncodedStrategy
    Strategy = Strategy

    ########################################################################
    ## market information
    def mHasLiquidityByPair(self, pair, AMMsells):
        """
        whether a specfic pair has any open liquidity positions

        :pair:          pair (as CarbonPair object)
        :AMMsells:      the token the AMM sells (as token name)
        :returns:       whether the directed pair has liquidity (boolean)
        """
        T = self._assertTokensProvided()
        pair=CarbonPair(pair)
        targetToken = AMMsells
        sourceToken = pair.other(targetToken)
        targetTokenO = T[targetToken]
        sourceTokenO = T[sourceToken]
        hasliqudity = self.hasLiquidityByPair(sourceTokenO.a, targetTokenO.a)
        return hasliqudity

    def mHasLiquidityByPairs(self, pairs):
        """
        whether a list of pairs has any open liquidity positions in any direction

        :pairs:     list of pairs (as CarbonPair objects or slashpair strings),
                    or a string of slashpairs separeted by commas 
        """
        T = self._assertTokensProvided()
        pairList = []
        pairListMeta = []
        if isinstance(pairs, str):
            pairs = [s.strip() for s in pairs.split(",")]
        for pair in pairs:
            if isinstance(pair, str):
                pair = CarbonPair(pair)
            targetTokenO = T[pair.tknq] # sic target = ...tknq
            sourceTokenO = T[pair.tknb]
            pairListMeta += [{"pair": pair, "AMMsells": targetTokenO.T}]
            pairList += [{"sourceToken": sourceTokenO.a, "targetToken": targetTokenO.a}]
            targetTokenO = T[pair.tknb]
            sourceTokenO = T[pair.tknq]
            pairListMeta += [{"pair": pair, "AMMsells": targetTokenO.T}]
            pairList += [{"sourceToken": sourceTokenO.a, "targetToken": targetTokenO.a}]
        result = self.hasLiquidityByPair(pairs=pairList)
        return tuple({**m, "hasLiquidity": r} for m, r in zip(pairListMeta, result))

    def mGetLiquidityByPair(self, pair, AMMsells):
        """
        liquidity values for multiple pairs

        :pair:          pair (as CarbonPair object or slashpair string)
        :AMMsells:      the token the AMM sells (as token name)
        :returns:       RangeByPair object
        """
        T = self._assertTokensProvided()
        pair = CarbonPair(pair)
        targetToken = AMMsells
        sourceToken = pair.other(targetToken)
        targetTokenO = T[targetToken]
        sourceTokenO = T[sourceToken]
        amount = self.getLiquidityByPair(sourceTokenO.a, targetTokenO.a)
        return self.PairLiquidity(
            amount=amount, 
            unit=targetToken, 
            AMMsells=AMMsells, 
            bidAsk="ask" if pair.has_basetoken(AMMsells) else "bid",
            pair=pair,
        )

    def mGetRangeByPair(self, pair, AMMsells):
        """
        gets the minimum and maximum exchange rate offered on Carbon between those two tokens

        :pair:          pair (as CarbonPair object or slashpair string)
        :AMMsells:      the token the AMM sells (as token name)
        :returns:       RangeByPair object
        """
        T = self._assertTokensProvided()
        pair = CarbonPair(pair)
        targetToken = AMMsells
        sourceToken = pair.other(targetToken)
        sourceTokenO = T[sourceToken]
        targetTokenO = T[targetToken]
        minRateRaw = self.getMinRateByPair(sourceTokenO.a, targetTokenO.a)
        maxRateRaw = self.getMaxRateByPair(sourceTokenO.a, targetTokenO.a)
        return self.RangeByPair(
            startRate = pair.price(price0=maxRateRaw, tknb0=sourceTokenO.T, tknq0=targetTokenO.T),
            endRate   = pair.price(price0=minRateRaw, tknb0=sourceTokenO.T, tknq0=targetTokenO.T),
            AMMbuys   = sourceTokenO.T,
            AMMsells  = targetTokenO.T,
            bidAsk    = "ask" if pair.has_basetoken(AMMsells) else "bid",
            pair      = pair, 
            inverted  = pair.has_quotetoken(targetTokenO.T), 
        )

    def mGetRateLiquidityDepthByPair(self, rate, pair, AMMsells):
        """
        liquidity depth (ie cumulative liquidity) at a given exchange rate

        :rate:          rate at which liquidity is evaluated (in units of the pair)
                        if the rate is an Array then the liquidity is evaluated at each rate
        :pair:          pair (as CarbonPair object or slashpair string)
        :AMMsells:      the token the AMM sells (as token name)
        :returns:       RangeByPair object; if rate is an Array so is the amount
        """
        T = self._assertTokensProvided()
        pair = CarbonPair(pair)
        targetToken = AMMsells
        sourceToken = pair.other(targetToken)
        sourceTokenO = T[sourceToken]
        targetTokenO = T[targetToken]
        inverted = pair.has_basetoken(targetTokenO.T)
        try:
            rate_raw = [1/r if inverted else r for r in rate]
            islist = True
        except:
            rate_raw = 1/rate if inverted else rate
            islist = False
        amount = self.getRateLiquidityDepthByPair(sourceTokenO.a, targetTokenO.a, rate_raw)
        return self.PairLiquidity(
            amount=amount, 
            unit=targetToken, 
            AMMsells=AMMsells, 
            bidAsk="ask" if pair.has_basetoken(AMMsells) else "bid",
            pair=pair,
            rate=rate,
            inverted=inverted,
            islist=islist,
        )
    
    def mGetMarketByPair(self, pair):
        """
        summarises market information for a given pair
        
        :pair:          pair (as CarbonPair object or slashpair string)
        :returns:       MarketByPair object
        """
        T = self._assertTokensProvided()
        pair = CarbonPair(pair)
        bidRange = self.mGetRangeByPair(pair, pair.tknq)
        askRange = self.mGetRangeByPair(pair, pair.tknb)
        bidLiq = self.mGetLiquidityByPair(pair, pair.tknq)
        askLiq = self.mGetLiquidityByPair(pair, pair.tknb)
        
        return self.MarketByPair(
            pair=pair.slashpair,
            bestBid=bidRange.startRate,
            lastBid=bidRange.endRate,
            liqBid=bidLiq.amount,
            bestAsk=askRange.startRate,
            lastAsk=askRange.endRate,
            liqAsk=askLiq.amount,
        )
    
    ########################################################################
    ## user information
    def mGetUserStrategies(self, user, nsd=True):
        """
        gets the strategies that are owned by the user

        :user:      user (as address)
        :returns:   a list of modified strategy objects
        :nsd:       number significant digits (True=default; None=no rounding)
        
        """
        strategies_raw = self.getUserStrategies(user)
        strategies = (self.reformatStrategy(s, nsd=nsd) for s in strategies_raw)
        strategies = (self.Strategy.from_sdk(s) for s in strategies)
        return tuple(strategies)


    ########################################################################
    ## maker actions
    def mCreateStrategy(self, pair, sellRangeStart, sellRangeEnd, sellAmountTknB, buyRangeStart, buyRangeEnd, buyAmountTknQ, overrides):
        """
        creates a strategy buying and selling tokens on a given pair


        :pair:                  pair (as CarbonPair object or slashpair string)
        :sellRangeStart:        minimum exchange rate at which to sell (in units of the pair)*
        :sellRangeEnd:          maximum exchange rate at which to sell (in units of the pair)*
        :sellAmountTknB:        amount to be sold (in units of base token)
        :buyRangeStart:         minimum exchange rate at which to buy (in units of the pair)*
        :buyRangeEnd:           maximum exchange rate at which to buy (in units of the pair)*
        :buyAmountTknQ:         amount to be bought (in units of quote token)

        *sell and buy in this case refers to the base token of the pair; maker orders are are buy low, 
        sell high, so the buy range -- in the natural quote direction of the pair -- is below the sell 
        range; selling starts at the low price and buying starts at the high price, so usually the order
        of the ranges is 
            
            buyRangeEnd < buyRangeStart <= sellRangeStart < sellRangeEnd
        """
        T = self._assertTokensProvided()
        pair = CarbonPair(pair)

        if not sellRangeStart <= sellRangeEnd:
            sellRangeStart, sellRangeEnd = sellRangeEnd, sellRangeStart
            print("[mCreateStrategy] sellRangeStart > sellRangeEnd, swapping", sellRangeStart, sellRangeEnd)

        if not buyRangeStart >= buyRangeEnd:
            buyRangeStart, buyRangeEnd = buyRangeEnd, buyRangeStart
            print("[mCreateStrategy] buyRangeStart < buyRangeEnd, swapping", buyRangeStart, buyRangeEnd)

        if not sellRangeStart > buyRangeStart:
            raise ValueError("sellRangeStart <= buyRangeStart", sellRangeStart, buyRangeStart)
        
        data = self.createBuySellStrategy(
            baseToken = T.ETH.a, 
            quoteToken = T.USDC.a, 
            buyPriceLow = buyRangeEnd, 
            buyPriceHigh = buyRangeStart, 
            buyBudget = buyAmountTknQ,
            sellPriceLow = sellRangeStart, 
            sellPriceHigh = sellRangeEnd, 
            sellBudget = sellAmountTknB,
            overrides = overrides, 
        )
        return data


    ########################################################################
    ## taker actions
    def mPrepareTrade(self, pair, traderBuySell, tkn, amount):
        """
        prepares a trade, calling getTradeData

        :pair:              pair (as CarbonPair object or slashpair string)
        :traderBuySell:     cls.BUY or cls.SELL, from PoV of the trader
        :tkn:               token to be bought or sold (as token name)
        :amount:            amount to be bought or sold (in units of tkn)
        :returns:           a TradeData object
        """
        T = self._assertTokensProvided()
        pair = CarbonPair(pair)
        if not traderBuySell in (self.BUY, self.SELL):
            raise ValueError("traderBuySell must be cls.BUY or cls.SELL", traderBuySell, self.BUY, self.SELL)
        if not pair.has_token(tkn):
            raise ValueError("tkn must be in pair", tkn, pair)
        if not amount > 0:
            raise ValueError("amount must be positive", amount)
        sourceToken = tkn if traderBuySell == self.SELL else pair.other(tkn)
        targetToken = pair.other(sourceToken)
        sourceTokenO = T[sourceToken]
        targetTokenO = T[targetToken]
        tradeByTargetAmount = (tkn == targetToken)
        if not Tokens.containsall({sourceToken, targetToken}):
            raise ValueError("not all token addresses are availabe", pair, Tokens)
        data = self.getTradeData(
            sourceToken=sourceTokenO.a, 
            targetToken=targetTokenO.a, 
            tradeByTargetAmount=tradeByTargetAmount, 
            amount=amount, 
        )
        tradeActions = [self.bn2intd(d) for d in data["tradeActions"]]
        actionsWei = [self.bn2intd(d) for d in data["actionsWei"]]
        actionsTokenRes = [self.bn2intd(self.str2floatd(d)) for d in data["actionsTokenRes"]]
        del data["tradeActions"]
        del data["actionsWei"]
        del data["actionsTokenRes"]
        tradeData =self.str2floatd(data)
        result = self.TradeData(
            pair=pair.slashpair,
            tkn=tkn,
            traderBuySell=traderBuySell,
            amount=amount,
            sourceToken=sourceTokenO,
            targetToken=targetTokenO,
            tradeByTargetAmount=tradeByTargetAmount,
            tradeActions = tradeActions,
            actionsWei = actionsWei,
            actionsTokenRes = actionsTokenRes,
            tradeData=tradeData,
        )
        return result
    
    MAXDEVIATIONPC = 0.1
    def mComposeTradeTransaction(self, sourceToken, targetToken, tradeActions, tradeData, tradeByTargetAmount, maxDeviationPc=None, deadline=None, overrides=None):
        """
        compose a trade transaction mostly based on the results of mPrepareTrade

        :sourceToken:               as returned by mPrepareTrade*
        :targetToken:               as returned by mPrepareTrade*
        :tradeActions:              as returned by mPrepareTrade*
        :tradeData:                 as returned by mPrepareTrade*
        :tradeByTargetAmount:       as returned by mPrepareTrade*
        :deadline:                  deadline (TODO)
        :maxDeviationPc:            max deviation from expected trade outcome in pc (default: 0.1 = 10%)**
        :overrides:                 overrides (TODO)
        :returns:                   a transaction object
        
        *NOTE: used cttargs or cttkwargs on the TradeData object to generate those
        **depending on the side specified this either determines a minimum amount received
        or a maximum amount delivered; this is expressed as percentage of the amount in tradeData
        """

        if maxDeviationPc is None: maxDeviationPc = self.MAXDEVIATIONPC
        if tradeByTargetAmount:
            data = self.composeTradeByTargetTransaction(
                    sourceToken=sourceToken.a, 
                    targetToken=targetToken.a, 
                    tradeActions=tradeActions, 
                    deadline=deadline,
                    maxInput=tradeData["totalSourceAmount"]*(1+maxDeviationPc),
                    overrides=overrides,
            )
        else:
            data = self.composeTradeBySourceTransaction(
                    sourceToken=sourceToken.a, 
                    targetToken=targetToken.a, 
                    tradeActions=tradeActions, 
                    deadline=deadline,
                    minReturn=tradeData["totalTargetAmount"]*(1-maxDeviationPc),
                    overrides=overrides,
            )
        return data

    

    ########################################################################
    ## helpers
    @classmethod
    def reformatOrders(cls, orders):
        """
        converts orders dict from BigNumber to int
        """
        return {k: cls.bn2intd(v) for k, v in orders.items()}


# ------------------------------------------------------------
# Auto generated test file `test_056_CarbonPair2.py`
# ------------------------------------------------------------
# source file   = NBTest_056_CarbonPair2.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 056
# test comment  = CarbonPair2
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, CarbonOrderUI, CarbonPair, __version__, __date__
from carbon.helpers import SharedVar, print_version
from math import floor, ceil, trunc
print("{0.__name__} {0.__VERSION__} ({0.__DATE__})".format(CarbonPair))
print_version(require="2.3.2")




# ------------------------------------------------------------
# Test      056
# File      test_056_CarbonPair2.py
# Segment   Ensure capitalisation
# ------------------------------------------------------------
def test_ensure_capitalisation():
# ------------------------------------------------------------
    
    assert CarbonPair("usd/eth").tknb == "USD"
    assert CarbonPair("usd/eth").tknq == "ETH"
    assert CarbonPair("usd/eth").slashpair == "USD/ETH"
    assert CarbonPair("bnBNT/eth").tknb == "BNBNT"
    
    assert CarbonPair(tknb="usd", tknq="eth").tknb == "USD"
    assert CarbonPair(tknb="usd", tknq="eth").tknq == "ETH"
    assert CarbonPair(tknb="usd", tknq="eth").slashpair == "USD/ETH"
    assert CarbonPair(tknb="bnBNT", tknq="eth").tknb == "BNBNT"
    

# ------------------------------------------------------------
# Test      056
# File      test_056_CarbonPair2.py
# Segment   Ensure originals are preserved
# ------------------------------------------------------------
def test_ensure_originals_are_preserved():
# ------------------------------------------------------------
    
    assert CarbonPair("usd/eth").tknb_o == "usd"
    assert CarbonPair("usd/eth").tknq_o == "eth"
    assert CarbonPair("usd/eth").slashpair_o == "usd/eth"
    assert CarbonPair("bnBNT/eth").tknb_o == "bnBNT"
    assert CarbonPair("bnBNT/eth").slashpair_o == "bnBNT/eth"
    
    assert CarbonPair(tknb="usd", tknq="eth").tknb_o == "usd"
    assert CarbonPair(tknb="usd", tknq="eth").tknq_o == "eth"
    assert CarbonPair(tknb="usd", tknq="eth").slashpair_o == "usd/eth"
    assert CarbonPair(tknb="bnBNT", tknq="eth").tknb_o == "bnBNT"
    assert CarbonPair(tknb="bnBNT", tknq="eth").slashpair_o == "bnBNT/eth"
    

# ------------------------------------------------------------
# Test      056
# File      test_056_CarbonPair2.py
# Segment   Ensure display values are correct
# ------------------------------------------------------------
def test_ensure_display_values_are_correct():
# ------------------------------------------------------------
    
    # ### display_orig = Default
    
    assert CarbonPair("usd/eth").tknb_d == "USD"
    assert CarbonPair("usd/eth").tknq_d == "ETH"
    assert CarbonPair("usd/eth").slashpair_d == "USD/ETH"
    assert CarbonPair("bnBNT/eth").tknb_d == "BNBNT"
    
    assert CarbonPair(tknb="usd", tknq="eth").tknb_d == "USD"
    assert CarbonPair(tknb="usd", tknq="eth").tknq_d == "ETH"
    assert CarbonPair(tknb="usd", tknq="eth").slashpair_d == "USD/ETH"
    assert CarbonPair(tknb="bnBNT", tknq="eth").tknb_d == "BNBNT"
    
    
    # ### display_orig = False
    
    assert CarbonPair("usd/eth", display_orig=False).tknb_d == "USD"
    assert CarbonPair("usd/eth", display_orig=False).tknq_d == "ETH"
    assert CarbonPair("usd/eth", display_orig=False).slashpair_d == "USD/ETH"
    assert CarbonPair("bnBNT/eth", display_orig=False).tknb_d == "BNBNT"
    
    assert CarbonPair(tknb="usd", tknq="eth", display_orig=False).tknb_d == "USD"
    assert CarbonPair(tknb="usd", tknq="eth", display_orig=False).tknq_d == "ETH"
    assert CarbonPair(tknb="usd", tknq="eth", display_orig=False).slashpair_d == "USD/ETH"
    assert CarbonPair(tknb="bnBNT", tknq="eth", display_orig=False).tknb_d == "BNBNT"
    
    
    # ### display_orig = True
    
    assert CarbonPair("usd/eth", display_orig=True).tknb_d == "usd"
    assert CarbonPair("usd/eth", display_orig=True).tknq_d == "eth"
    assert CarbonPair("usd/eth", display_orig=True).slashpair_d == "usd/eth"
    assert CarbonPair("bnBNT/eth", display_orig=True).tknb_d == "bnBNT"
    assert CarbonPair("bnBNT/eth", display_orig=True).slashpair_d == "bnBNT/eth"
    
    assert CarbonPair(tknb="usd", tknq="eth", display_orig=True).tknb_d == "usd"
    assert CarbonPair(tknb="usd", tknq="eth", display_orig=True).tknq_d == "eth"
    assert CarbonPair(tknb="usd", tknq="eth", display_orig=True).slashpair_d == "usd/eth"
    assert CarbonPair(tknb="bnBNT", tknq="eth", display_orig=True).tknb_d == "bnBNT"
    assert CarbonPair(tknb="bnBNT", tknq="eth", display_orig=True).slashpair_d == "bnBNT/eth"
    
    


# ------------------------------------------------------------
# Auto generated test file `test_905_ReconciliationsSlow.py`
# ------------------------------------------------------------
# source file   = NBTest_905_ReconciliationsSlow.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 905
# test comment  = ReconciliationsSlow
# ------------------------------------------------------------



from carbon.helpers.stdimports import *
from carbon.helpers.soltest import SolTestBase
from carbon.helpers.floatint import *
#print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(SolTestBase))
#print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonFloatInt32))
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonOrderUI))
#print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(P))
from math import log2, floor, ceil, sqrt
print_version(require="2.3.3")




# ------------------------------------------------------------
# Test      905
# File      test_905_ReconciliationsSlow.py
# Segment   Reconcile CarbonOrderUI vs Core Sim
# ------------------------------------------------------------
def test_reconcile_carbonorderui_vs_core_sim():
# ------------------------------------------------------------
    
    Sim = CarbonSimulatorUI(pair=P("ETH/USDC"), raiseonerror=True)
    r = Sim.add_strategy("ETH", 10, 1500, 2000, 1000, 750, 500)
    r["orders"]
    
    ouis=r["orderuis"]
    ouis
    
    # ### amm_sells ETH (dx from dy)
    
    size = 0.1
    r = Sim.amm_sells("ETH", size, execute=False)
    tr = r["trades"].query("aggr==True").iloc[0]
    r = Sim.amm_sells("ETH", size, execute=False)
    tr2 = r["trades"].query("aggr==True").iloc[0]
    assert tr["price"] == tr2["price"] # check trades are NOT being executed
    dx = ouis[0].dxfromdy_f(size)
    assert abs(dx/tr["amt2"] - 1) < 1e-10
    tr2["amt2"], dx
    
    for size in np.linspace(0.1,1,10):
        r = Sim.amm_sells("ETH", size, execute=False)
        tr = r["trades"].query("aggr==True").iloc[0]
        amts = tr["amt2"]
        dx = ouis[0].dxfromdy_f(size)
        print(size, amts, dx, amts/dx-1)
        assert abs(amts/dx - 1) < 1e-10
    
    # ### amm_sells USDC (dx from dy)
    
    for size in np.linspace(0.1,1000,10):
        r = Sim.amm_sells("USDC", size, execute=False)
        tr = r["trades"].query("aggr==True").iloc[0]
        amts = tr["amt2"]
        dx = ouis[1].dxfromdy_f(size)
        print(size, amts, dx, amts/dx-1)
        assert abs(amts/dx - 1) < 1e-8
    
    # ### amm_buys USDC (dy from dx)
    
    for size in np.linspace(0.1,1000,10):
        r = Sim.amm_buys("USDC", size, execute=False)
        tr = r["trades"].query("aggr==True").iloc[0]
        amts = tr["amt1"]
        dx = ouis[0].dyfromdx_f(size)
        print(size, amts, dx, amts/dx-1)
        assert abs(amts/dx - 1) < 1e-8
    
    # ### amm_buys ETH (dy from dx)
    
    for size in np.linspace(0.1,1,10):
        r = Sim.amm_buys("ETH", size, execute=False)
        tr = r["trades"].query("aggr==True").iloc[0]
        amts = tr["amt1"]
        dx = ouis[1].dyfromdx_f(size)
        print(size, amts, dx, amts/dx-1)
        assert abs(amts/dx - 1) < 1e-8
    

# ------------------------------------------------------------
# Test      905
# File      test_905_ReconciliationsSlow.py
# Segment   Verifying equivalence of both order sets
# ------------------------------------------------------------
def test_verifying_equivalence_of_both_order_sets():
# ------------------------------------------------------------
    
    Sim = CarbonSimulatorUI(pair=P("ETH/USDC"), raiseonerror=True)
    Sim.add_strategy("ETH", 1, 1500, 2000, 1000, 750, 500)
    Sim.add_strategy("ETH", 1.1, 1501, 2001, 1001, 751, 501)
    r = Sim.state()
    os = r["orders"]
    os
    
    ouis = r["orderuis"]
    ouis
    
    assert len(ouis)==len(os)
    for oui, oo in zip(ouis.values(),os.iterrows()):
        o = oo[1]
        #print (oui, o)
        assert o["id"] == oui.id
        assert o["lid"] == oui.lid
        assert o["pair"] == oui.pair.pair_iso
        assert o["tkn"] == oui.tkn
        assert o["tkn"] == oui.tkn
        assert o["y_int"] == oui.yint
        assert abs(o["p_start"]/oui.pa-1) < 1e-10
        assert abs(o["p_end"]/oui.pb-1) < 1e-10
        assert abs(o["y_int"]/oui.yint-1) < 1e-10
        assert abs(o["y"]/oui.y-1) < 1e-10
        assert o["p_unit"] == oui.price_convention
        assert oui.linked.linked is oui
    
    


"""
Carbon helper module - read time series from data frame
"""
__VERSION__ = "2.2"
__DATE__ = "27/01/2023"

import pandas as _pd

# those imports are for re-export
from os.path import join as j
import os

def tknpair(slashpair, invert=False):
    """
    returns (tknb, tnkq) is pair is slashpaur, else ("RSK", "CSH") (or inverted)
    """
    split = slashpair.split("/")
    if len(split) == 2:
        result = split[0].strip(), split[1].strip()
    else:
        result = ("RSK", "CSH")
    return result if not invert else tuple(result[::-1])

def pdread(
        fn, 
        datacol=None, 
        indexcol=None, 
        from_ts=None, 
        from_pc=None, 
        period_days=None, 
        period_pc=None, 
        min_dt=None,
        invert=False,
        tkns=False,
    ):
    """
    reads a dataframe and returns a single column with index
    
    :fn:            the (full) filename
    :datacol:       name or index of the data col; None returns frame
    :indexcol:      name of the index col (default: "datetime")
    :from_ts:       the start of the observation period, in a formate that can be converted
                    by pd.Timestamp(from_ts); default is beginning of series
    :from_pc:       alternative to from_ts; sets the start date as perc of full time
    :period_days:   the observation period in days (inclusive); default end of series
    :period_pc:     alternative to period_days; sets period as perc of full time
    :min_dt:        before anything else is applied, everything before min_dt is discarded
    :invert:        if True, invert the data series (raises if datacol is None)
    :tkns:          if True, also return token names in case of series
    :returns:       pandas series (or data frame if indexcol is None)
    """
    if indexcol is None: indexcol = "datetime"

    df = _pd.read_pickle(fn)
    if min_dt: df = df[df.index >= min_dt]
    
    if not period_days is None and not period_pc is None:
        raise ValueError("Not both period_days and period_pc can not-None")
    if not period_pc is None:
        period_days = period_pc * (df.index[-1]-df.index[0])/_pd.Timedelta(days=1)
        #print("[pdread]", period_days)

    if not from_ts is None and not from_pc is None:
        raise ValueError("Not both from_ts and from_pc can not-None")
    if not from_pc is None:
        from_ts = df.index[0] + from_pc * (df.index[-1]-df.index[0])

    if not from_ts is None or not period_days is None:
        if from_ts is None:
            from_ts = df.index[0]
        else:
            from_ts = _pd.Timestamp(from_ts)
            if from_ts < df.index[0]:
                from_ts = df.index[0]
        if period_days is None:
            to_ts = df.index[-1]
        else:
            to_ts = from_ts + period_days * _pd.Timedelta(days=1)
        to_ts   = _pd.Timestamp(to_ts)
    else:
        from_ts = df.index[0]
        to_ts = df.index[-1]
    
    #print("[pdread]", from_ts, to_ts)
    df = df[(df.index >= from_ts) & (df.index <= to_ts)]
    
    if datacol is None:
        if invert:
            raise ValueError("Can't use invert when returning a whole dataframe", fn, indexcol)
        return df.copy()
    elif isinstance(datacol, str):
        result = df[datacol].copy()
        if tkns:
            return (result if not invert else 1/result, tknpair(datacol, invert))
        else:
            return result if not invert else 1/result
    elif isinstance(datacol, int):
        result = df.iloc[:, datacol].copy()
        if tkns:
            return (result if not invert else 1/result, tknpair("RSK/CSH", invert))
        else:
            return result if not invert else 1/result
    else:
        raise ValueError("datacol must be None, str or int", datacol)

def pdcols(fn):
    """
    reads a dataframe and returns a single column with index
    
    :fn:        the (full) filename
    :returns:   the column names (excluding the first two)
    """
    return _pd.read_pickle(fn).columns

def pathtime(path):
    """
    DEPRECATED returns the time (in years) covered by the series `path`
    
    use pathtime_yrs
    """
    return (path.index[-1]-path.index[0])/_pd.Timedelta(days=1)/365.25

def pathtime_yrs(path):
    """
    returns the time (in years) covered by the series `path`

    NOTE: conversion factor is 365.25
    """
    return pathtime_days(path)/365.25

def pathtime_days(path):
    """
    returns the time (in days) covered by the series `path`
    """
    return (path.index[-1]-path.index[0])/_pd.Timedelta(days=1)


# ------------------------------------------------------------
# Auto generated test file `test_055_PathInterpolation.py`
# ------------------------------------------------------------
# source file   = NBTest_055_PathInterpolation.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 055
# test comment  = PathInterpolation
# ------------------------------------------------------------



from carbon import CarbonSimulatorUI, CarbonOrderUI, P, __version__, __date__
from carbon.helpers import print_version, PathInterpolation as PI, interpolate, plt_style
from matplotlib import pyplot as plt
import pandas as pd

plt.rcParams['figure.figsize'] = [12,6]
plt_style('seaborn-v0_8-dark', 'seaborn-dark')
print_version(require="2.2.6")




# ------------------------------------------------------------
# Test      055
# File      test_055_PathInterpolation.py
# Segment   Path Interpolation [NOTEST]
# ------------------------------------------------------------
def notest_path_interpolation():
# ------------------------------------------------------------
    
    # ### Gaussian bridge
    
    y = PI.gaussian_bridge(100, 100, 0.3, 1000)
    plt.plot(y)
    
    PI.estimate_vol_np(y)
    
    # ### Path segment
    
    path_segment = pd.Series([100,120], index=[pd.Timestamp("2020-01-01"), pd.Timestamp("2020-07-01")])
    path_segmenti = PI.interpolate_segment(path_segment, sig=0.5, period=PI.days(1))
    plt.plot(path_segment)
    plt.plot(path_segmenti)
    print("estimated vol:", PI.estimate_vol(path_segmenti))
    path_segment
    
    # ### Path
    
    path_data = [100,130,110,95,125,155,120]
    numseg = len(path_data)-1
    start_dt = pd.Timestamp("2020-01-01")
    end_dt = pd.Timestamp("2020-07-01")
    delt = (end_dt-start_dt)/numseg
    path_ix = [start_dt+i*delt for i in range(numseg+1)]
    path = pd.Series(path_data, index=path_ix)
    pathi = PI.interpolate(path, sig=0.5, period=PI.hours(1))
    plt.plot(path)
    plt.plot(pathi)
    print("estimated vol:", PI.estimate_vol(pathi))
    

# ------------------------------------------------------------
# Test      055
# File      test_055_PathInterpolation.py
# Segment   Path Interpolation
# ------------------------------------------------------------
def test_path_interpolation():
# ------------------------------------------------------------
    
    assert PI.minutes(60) == PI.hours(1)
    assert PI.days(1) == PI.hours(24)
    
    path_segment = pd.Series([100,120], index=[pd.Timestamp("2020-01-01"), pd.Timestamp("2020-07-01")])
    path_segmenti = PI.interpolate_segment(path_segment, sig=0.5, period=PI.days(1))
    assert len(path_segmenti) == 183
    assert path_segmenti[0] == 100
    assert path_segmenti[-1] == 120
    assert path_segmenti.index[1]-path_segmenti.index[0] == PI.days(1)
    assert PI.estimate_vol(path_segmenti)
    
    path_data = [100,130,110,95,125,155,120]
    numseg = len(path_data)-1
    start_dt = pd.Timestamp("2020-01-01")
    end_dt = pd.Timestamp("2020-02-01")
    delt = (end_dt-start_dt)/numseg
    path_ix = [start_dt+i*delt for i in range(numseg+1)]
    path = pd.Series(path_data, index=path_ix)
    pathi = PI.interpolate(path, sig=0.5, period=PI.hours(1))
    
    assert len(pathi)==745
    assert pathi.index[1]-pathi.index[0] == PI.hours(1)
    assert pathi[0] == 100
    assert pathi[-1] == 120
    
    
    


# ------------------------------------------------------------
# Auto generated test file `test_062_CarbonOrderUI.py`
# ------------------------------------------------------------
# source file   = NBTest_062_CarbonOrderUI.py
# source path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# target path   = /Users/skl/REPOES/Bancor/CarbonSimulator/resources/NBTest/
# test id       = 062
# test comment  = CarbonOrderUI
# ------------------------------------------------------------



from carbon.helpers.stdimports import *
from carbon import CarbonOrderUI
print("{0.__name__} v{0.__VERSION__} ({0.__DATE__})".format(CarbonOrderUI))
print_version(require="2.3.3")


#


# ------------------------------------------------------------
# Test      062
# File      test_062_CarbonOrderUI.py
# Segment   Selling the base token
# ------------------------------------------------------------
def test_selling_the_base_token():
# ------------------------------------------------------------
    
    o = CarbonOrderUI.from_prices(
        pair="ETH/USDC", 
        tkn="ETH", 
        pa=2500, 
        pb=3000, 
        yint=10, 
        y=5)
    o
    
    # ### Prices
    #
    # Below the different ways of accessing the prices. `pa=py=p_start` is always the beginning of the range, and `pb=px=p_end` is always the end of the range. `pmax` and `pmin` are always the higher and lower price in the quote convention of the pair, respectively. Those figures are quoted in the convention of the pair, so as we are selling the base token this is the inverse of the internal quotation $dy/dx$ which is `pa_raw` and `pb_raw`. `p0` and `p0_raw` are is the geometric average of the start and end prices in the respective quotation. `reverseq` is true if and only if the `_raw` figures are the inverse of the other figures.
    
    print("p_start", o.p_start)
    print("p_end", o.p_end)
    print("p0", o.p0)
    print("reverseq", o.reverseq)
    print("pa_raw", o.pa_raw)
    print("pb_raw", o.pb_raw)
    print("p0_raw", o.p0_raw)
    print("p_min", o.pmin)
    print("p_max", o.pmax)
    assert abs(o.p_start/2500-1)<1e-10
    assert abs(o.p_end/3000-1)<1e-10
    assert o.pa is o.p_start
    assert o.pb is o.p_end
    assert o.py is o.pa
    assert o.px is o.pb
    assert o.pmin == min(o.p_start, o.p_end)
    assert o.pmax == max(o.p_start, o.p_end)
    assert o.pa_raw == 1/o.pa if o.reverseq else o.pa
    assert o.pb_raw == 1/o.pb if o.reverseq else o.pb
    assert abs(o.p0/sqrt(o.pa*o.pb)-1)<1e-10
    assert abs(o.p0_raw/sqrt(o.pa_raw*o.pb_raw)-1)<1e-10
    
    # ### Curve capacity
    #
    # `yint` is the curve capacity in the token being sold (here: ETH), and `xint` the corresponding sale amount. `y0` and `x0` are reference parameters determining the curve, and we have `yint/xint=y0/x0=p0_raw`
    
    print("yint", o.yint)
    print("xint", o.xint)
    print("y0", o.y0)
    print("x0", o.x0)
    assert o.yint == 10
    assert abs(o.yint/o.xint/o.p0_raw-1)<1e-10
    assert o.y0/o.x0 == o.p0_raw
    
    # ### Curve convexity
    #
    # The curve convexity determines how different `pa` and `pb` are. There are a number of ultimately equivalent paramters that describe that
    #
    # - `widthpc` is the percentage width, defined as `(pa_raw-pb_raw)/p0_raw`
    # - `widthr` is the width ratio, defined as `pa_raw/pb_raw >= 1`
    # - `Gamma` is the curve leverage parameter between `0..1` where `0` corresponds to constant price and `1` corresponds to constant product
    # - `Q` is an alternative convexity parameter also between `0..1` but the other way
    #
    # The relationshipt between the parameters is provided in the functions `Gamma_from_Q`
    #
    # $$
    # \Gamma(Q) = 1 - \sqrt{Q}
    # $$ 
    #
    # and `Q_from_Gamma` 
    #
    # $$
    # Q(\Gamma) = (1-\Gamma)^2
    # $$
    #
    # Also `Q=1/sqrt(widthr)` because
    #
    # $$
    # Q = \sqrt{\frac{p_a}{p_b}}
    # $$
    
    print("widthpc", o.widthpc)
    print("widthr", o.widthr)
    print("Gamma", o.Gamma)
    print("Q", o.Q)
    assert abs(o.widthpc/((o.pa_raw-o.pb_raw)/o.p0_raw)-1)<1e-10
    assert abs(o.widthr/(o.pa_raw / o.pb_raw)-1)<1e-10
    assert abs(o.widthr/(o.pmax / o.pmin)-1)<1e-10
    assert abs(o.Q/(sqrt(1/o.widthr))-1)<1e-10
    assert abs(o.Q/o.Q_from_Gamma(o.Gamma)-1)<1e-10
    assert abs(o.Gamma/o.Gamma_from_Q(o.Q)-1)<1e-10
    
    # ### Asymptotes
    #
    # Here we deal with the curve asymptotes `yasym` and `xasym`. They are related to the other token figures via 
    #
    # $$
    # \frac{x_{asym}}{x_0} = \frac{y_{asym}}{y_0} = 1-\frac 1 \Gamma
    # $$
    #
    # We also look at the relationship between the `int`ercepts and the `0` figures which are
    #
    # $$
    # \frac{x_{int}}{x_0} = \frac{y_{int}}{y_0} = \frac {2-\Gamma} {1-\Gamma}
    # $$
    
    print("yasym", o.yasym)
    print("xasym", o.xasym)
    print("yint", o.yint)
    print("xint", o.xint)
    print("y0", o.y0)
    print("x0", o.x0)
    assert abs(o.xasym/o.x0/(o.yasym/o.y0)-1)<1e-10
    assert abs(o.yasym/o.y0/o.asym_over_0(o.Gamma)-1)<1e-10
    assert abs(o.xasym/o.x0/o.asym_over_0(o.Gamma)-1)<1e-10
    assert abs(o.xint/o.x0/(o.yint/o.y0)-1)<1e-10
    assert abs(o.yint/o.y0/o.int_over_0(o.Gamma)-1)<1e-10
    assert abs(o.xint/o.x0/o.int_over_0(o.Gamma)-1)<1e-10
    
    # ### Kappa and the invariant function
    #
    # A very elegant way to express the invariant function is 
    #
    # $$
    # (x-x_{asym})(y-y_{asym}) = \kappa = \frac{x_0y_0}{\Gamma^2}
    # $$
    #
    # or in way that scales better as
    #
    # $$
    # \sqrt{(x-x_{asym})(y-y_{asym})} = \bar\kappa = \frac{\sqrt{x_0y_0}}{\Gamma}
    # $$
    #
    
    print("y0", o.y0)
    print("x0", o.x0)
    print("Gamma", o.Gamma)
    print("leverage_fctr", o.leverage_fctr)
    print("kappa", o.kappa)
    print("kappa_bar", o.kappa_bar)
    assert o.leverage_fctr == 1/o.Gamma
    assert abs(o.kappa/(o.x0*o.y0/o.Gamma**2)-1)<1e-10
    assert abs(o.kappa_bar/(sqrt(o.x0*o.y0)/o.Gamma)-1)<1e-10
    
    # Here we are checking the invariant function against the expression above. We are comparing both against `xfromy_f` and `yfromx_f` where we recall that `y` is the token being sold and `x` the token being bought by the AMM.
    
    for y in [0.1,1,3,5,7,9,9.99]:
        x = o.xfromy_f(y)
        f2 = (x-o.xasym)*(y-o.yasym)
        f = sqrt(f2)
        print(f"y={y} ->", x, f, o.kappa_bar, f2, o.kappa)
        assert abs(f/o.kappa_bar-1) < 1e-10
        assert abs(f2/o.kappa-1) < 1e-10
    

# ------------------------------------------------------------
# Test      062
# File      test_062_CarbonOrderUI.py
# Segment   Selling the quote token
# ------------------------------------------------------------
def test_selling_the_quote_token():
# ------------------------------------------------------------
    
    o = CarbonOrderUI.from_prices(
        pair="ETH/USDC", 
        tkn="USDC", 
        pa=1500, 
        pb=1000, 
        yint=3000, 
        y=1500)
    o
    
    # ### Prices
    
    print("p_start", o.p_start)
    print("p_end", o.p_end)
    print("p0", o.p0)
    print("reverseq", o.reverseq)
    print("pa_raw", o.pa_raw)
    print("pb_raw", o.pb_raw)
    print("p0_raw", o.p0_raw)
    print("p_min", o.pmin)
    print("p_max", o.pmax)
    assert abs(o.p_start/1500-1)<1e-10
    assert abs(o.p_end/1000-1)<1e-10
    assert o.pa is o.p_start
    assert o.pb is o.p_end
    assert o.py is o.pa
    assert o.px is o.pb
    assert o.pmin == min(o.p_start, o.p_end)
    assert o.pmax == max(o.p_start, o.p_end)
    assert o.pa_raw == 1/o.pa if o.reverseq else o.pa
    assert o.pb_raw == 1/o.pb if o.reverseq else o.pb
    assert abs(o.p0/sqrt(o.pa*o.pb)-1)<1e-10
    assert abs(o.p0_raw/sqrt(o.pa_raw*o.pb_raw)-1)<1e-10
    
    # ### Curve capacity
    #
    
    print("yint", o.yint)
    print("xint", o.xint)
    print("y0", o.y0)
    print("x0", o.x0)
    assert o.yint == 3000
    assert abs(o.yint/o.xint/o.p0_raw-1)<1e-10
    assert o.y0/o.x0 == o.p0_raw
    
    # ### Curve convexity
    
    print("widthpc", o.widthpc)
    print("widthr", o.widthr)
    print("Gamma", o.Gamma)
    print("Q", o.Q)
    assert abs(o.widthpc/((o.pa_raw-o.pb_raw)/o.p0_raw)-1)<1e-10
    assert abs(o.widthr/(o.pa_raw / o.pb_raw)-1)<1e-10
    assert abs(o.widthr/(o.pmax / o.pmin)-1)<1e-10
    assert abs(o.Q/(sqrt(1/o.widthr))-1)<1e-10
    assert abs(o.Q/o.Q_from_Gamma(o.Gamma)-1)<1e-10
    assert abs(o.Gamma/o.Gamma_from_Q(o.Q)-1)<1e-10
    
    # ### Asymptotes
    
    print("yasym", o.yasym)
    print("xasym", o.xasym)
    print("yint", o.yint)
    print("xint", o.xint)
    print("y0", o.y0)
    print("x0", o.x0)
    assert abs(o.xasym/o.x0/(o.yasym/o.y0)-1)<1e-10
    assert abs(o.yasym/o.y0/o.asym_over_0(o.Gamma)-1)<1e-10
    assert abs(o.xasym/o.x0/o.asym_over_0(o.Gamma)-1)<1e-10
    assert abs(o.xint/o.x0/(o.yint/o.y0)-1)<1e-10
    assert abs(o.yint/o.y0/o.int_over_0(o.Gamma)-1)<1e-10
    assert abs(o.xint/o.x0/o.int_over_0(o.Gamma)-1)<1e-10
    
    # ### Kappa and invariant
    
    print("y0", o.y0)
    print("x0", o.x0)
    print("Gamma", o.Gamma)
    print("leverage_fctr", o.leverage_fctr)
    print("kappa", o.kappa)
    print("kappa_bar", o.kappa_bar)
    assert o.leverage_fctr == 1/o.Gamma
    assert abs(o.kappa/(o.x0*o.y0/o.Gamma**2)-1)<1e-10
    assert abs(o.kappa_bar/(sqrt(o.x0*o.y0)/o.Gamma)-1)<1e-10
    
    # Here we are checking the invariant function against the expression above. We are comparing both against `xfromy_f` and `yfromx_f` where we recall that `y` is the token being sold and `x` the token being bought by the AMM.
    
    for y in [1,100,1000,1500,2500,2999]:
        x = o.xfromy_f(y)
        f2 = (x-o.xasym)*(y-o.yasym)
        f = sqrt(f2)
        print(f"y={y} ->", x, f, o.kappa_bar, f2, o.kappa)
        assert abs(f/o.kappa_bar-1) < 1e-10
        assert abs(f2/o.kappa-1) < 1e-10
    
    
    


"""
Carbon -- representing a levered constant product curve

(c) Copyright Bprotocol foundation 2022. 
Licensed under MIT

NOTE: this class is not part of the API of the Carbon protocol, and you must expect breaking
changes even in minor version updates. Use at your own risk.

v1.0: ConstantProductCurve class
v1.1: added CPCContainer class
"""
__VERSION__ = "1.1"
__DATE__ = "16/Mar/2023"

from dataclasses import dataclass, field
import random
from math import sqrt
import numpy as np

try:
    dataclass_ = dataclass(frozen=True, kw_only=True)
except:
    dataclass_ = dataclass(frozen=True)
    
@dataclass_
class ConstantProductCurve():
    """
    represents a, potentially levered, constant product curve
    
    :k:        pool constant k = xy [x=k/y, y=k/x]
    :x:        (virtual) pool state x (virtual number of base tokens for sale)
    :x_act:    actual pool state x (actual number of base tokens for sale)
    :y_act:    actual pool state y (actual number of quote tokens for sale)
    :pair:     token pair in slash notation ("TKNB/TKNQ")
    :cid:      unique id (optional)
    
    NOTE: use the alternative constructors `from_xx` rather then the canonical one 
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__
    
    k: float
    x: float
    x_act: float = None
    y_act: float = None
    pair: str = None
    cid: any=None
        
    def __post_init__(self):
        
        if self.x_act is None: 
            super().__setattr__('x_act', self.x) # required because class frozen
        
        if self.y_act is None: 
            super().__setattr__('y_act', self.y) # ditto
            
        if self.pair is None:
            super().__setattr__('pair', "TKNB/TKNQ")
            
        super().__setattr__('pair', self.pair.upper())
            
        if self.x_act > self.x:
            print("[ConstantProductCurve] x_act > x:", self.x_act, self.x)
            
        if self.y_act > self.y:
            print("[ConstantProductCurve] y_act > y:", self.y_act, self.y)
    
    def setcid(self, cid):
        """sets the curve id [can only be done once]"""
        assert self.cid is None, "cid can only be set once"
        super().__setattr__('cid', cid)
        return self

    @classmethod
    def from_kx(cls, k, x, x_act=None, y_act=None, pair=None):
        "constructor: from k,x (and x_act, y_act)"
        return cls(k=k, x=x, x_act=x_act, y_act=y_act, pair=pair)
   
    @classmethod
    def from_ky(cls, k, y, x_act=None, y_act=None, pair=None):
        "constructor: from k,y (and x_act, y_act)"
        return cls(k=k, x=k/y, x_act=x_act, y_act=y_act, pair=pair) 
    
    @classmethod
    def from_xy(cls, x, y, x_act=None, y_act=None, pair=None):
        "constructor: from x,y (and x_act, y_act)"
        return cls(k=x*y, x=x, x_act=x_act, y_act=y_act, pair=pair)
    
    @classmethod
    def from_pk(cls, p, k, x_act=None, y_act=None, pair=None):
        "constructor: from k,p (and x_act, y_act)"
        return cls(k=k, x=sqrt(k/p), x_act=x_act, y_act=y_act, pair=pair)
    
    @classmethod
    def from_px(cls, p, x, x_act=None, y_act=None, pair=None):
        "constructor: from x,p (and x_act, y_act)"
        return cls(k=x*x*p, x=x, x_act=x_act, y_act=y_act, pair=pair)
    
    @classmethod
    def from_py(cls, p, y, x_act=None, y_act=None, pair=None):
        "constructor: from y,p (and x_act, y_act)"
        return cls(k=y*y/p, x=y/p, x_act=x_act, y_act=y_act, pair=pair)
    
    @classmethod
    def from_pkpp(cls, p, k, p_min=None, p_max=None, pair=None):
        "constructor: from k, p, p_min, p_max (default for last two is p)"
        if p_min is None: p_min = p
        if p_max is None: p_max = p
        x0 = sqrt(k/p)
        y0 = sqrt(k*p)
        xa = x0 - sqrt(k/p_max)
        ya = y0 - sqrt(k*p_min)
        return cls(k=k, x=x0, x_act=xa, y_act=ya, pair=pair)
    
    @property
    def tknb(self):
        "base token"
        return self.pair.split("/")[0]
    tknx = tknb
    
    @property
    def tknq(self):
        "quote token"
        return self.pair.split("/")[1]
    tkny = tknq

    @property
    def descr(self):
        "description of the pool"
        s1 = f"tknx = {self.x_act} [virtual: {self.x}] {self.tknx}\n"
        s2 = f"tkny = {self.y_act} [virtual: {self.y}] {self.tkny}\n"
        s3 = f"p    = {self.p} [min={self.p_min}, max={self.p_max}] {self.tknq} per {self.tknb}"
        return s1+s2+s3
    
    @property
    def y(self):
        "(virtual) pool state x (virtual number of base tokens for sale)"
        return self.k/self.x
    
    @property
    def p(self):
        "pool price (in dy/dx)"
        return self.y/self.x

    @property
    def kbar(self):
        "kbar = sqrt(k); kbar scales linearly with the pool size"
        return sqrt(self.k)
    
    @property
    def x_min(self):
        "minimum (virtual) x value"
        return self.x - self.x_act
    
    @property
    def y_min(self):
        "minimum (virtual) y value"
        return self.y - self.y_act
    
    @property
    def x_max(self):
        "maximum (virtual) x value"
        if self.y_min > 0:
            return self.k/self.y_min
        else:
            return None
    
    @property
    def y_max(self):
        "maximum (virtual) y value"
        if self.x_min > 0:
            return self.k/self.x_min
        else:
            return None
    
    @property
    def p_max(self):
        "maximum pool price (in dy/dx; None if unlimited) = y_max/x_min"
        if not self.x_min is None and self.x_min > 0:
            return self.y_max/self.x_min
        else:
            return None
    
    @property
    def p_min(self):
        "minimum pool price (in dy/dx; None if unlimited) = y_min/x_max"
        if not self.x_max is None and self.x_max > 0:
            return self.y_min/self.x_max
        else:
            return None
        
    def yfromx_f(self, x, ignorebounds=False):
        "y value for given x value (if in range; None otherwise)"
        y = self.k/x
        if ignorebounds:
            return y
        if not self.inrange(y, self.y_min, self.y_max):
            return None
        return y
    
    def xfromy_f(self, y, ignorebounds=False):
        "x value for given y value (if in range; None otherwise)"
        x = self.k/y
        if ignorebounds:
            return x
        if not self.inrange(x, self.x_min, self.x_max):
            return None
        return x
    
    def dyfromdx_f(self, dx, ignorebounds=False):
        "dy value for given dx value (if in range; None otherwise)"
        y = self.yfromx_f(self.x + dx, ignorebounds=ignorebounds)
        if y is None:
            return None
        return y-self.y
    
    def dxfromdy_f(self, dy, ignorebounds=False):
        "dx value for given dy value (if in range; None otherwise)"
        x = self.xfromy_f(self.y + dy, ignorebounds=ignorebounds)
        if x is None:
            return None
        return x-self.x
    
    @staticmethod
    def inrange(v, minv=None, maxv=None):
        "True if minv <= v <= maxv; None means no boundary"
        if not minv is None:
            if v < minv:
                return False
        if not maxv is None:
            if v > maxv:
                return False
        return True
    

@dataclass
class CPCContainer():
    """
    container for ConstantProductCurve objects (use += to add items)
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    curves: list = field(default_factory=list, repr=False)
        
    def __post_init__(self):
        pass
    
    def add(self, item):
        """adds a single ConstantProductCurve item to the container"""
        assert isinstance(item, ConstantProductCurve)
        if item.cid is None:
            item.setcid(len(self))
        self.curves += [item]
        return self
    
    def __iadd__(self, other):
        """alias for either add"""
        return self.add(other)
    
    def __iter__(self):
        return iter(self.curves)
    
    def __len__(self):
        return len(self.curves)
    
    @property
    def tknys(self):
        """returns set of all base tokens used by the curves"""
        return {c.tknb for c in self.curves}
    
    @property
    def tknxs(self):
        """returns set of all quote tokens used by the curves"""
        return {c.tknq for c in self.curves}
    
    @property
    def tkns(self):
        """returns set of all tokens used by the curves"""
        return self.tknxs.union(self.tknys)
    
    @property
    def pairs(self):
        """returns set of all pairs"""
        return {c.pair for c in self}
    
    def bypair(self, pair):
        """returns all curves by (directed!) pair"""
        return tuple(c for c in self if c.pair==pair)
    
    def bytknx(self, tknx):
        """returns all curves by quote token tknx"""
        return tuple(c for c in self if c.tknq==tknx)
    
    def bytkny(self, tkny):
        """returns all curves by base token tkny"""
        return tuple(c for c in self if c.tknb==tkny)
    
    @staticmethod                      
    def u(minx, maxx):
        """helper: returns uniform random var"""
        return random.uniform(minx, maxx)
    
    @property                      
    def u1(self):
        """helper: returns uniform [0,1] random var"""
        return random.uniform(0, 1)
    
    @dataclass
    class xystatsd():
        mean: any
        minv: any
        maxv: any
        sdev: any
            
    def xystats(self, curves=None):
        """calculates mean, min, max, stdev of x and y"""
        if curves is None:
            curves = self.curves
        tknx = {c.tknq for c in curves}
        tkny = {c.tknb for c in curves}
        assert len(tknx)==1 and len(tkny)==1, "all curves must have same tknq and tknb"
        x = [c.x for c in curves]
        y = [c.y for c in curves]
        return (
            self.xystatsd(np.mean(x), np.min(x), np.max(x), np.std(x)), 
            self.xystatsd(np.mean(y), np.min(y), np.max(y), np.std(y))
        )
    
    @property
    def tokentable(self):
        """returns dict associating tokens with the curves on which they appeay"""
        return {tkn: {
            "x": [i for i,c in enumerate(self) if c.tknq == tkn], 
            "y": [i for i,c in enumerate(self) if c.tknb == tkn]
            }
            for tkn in self.tkns
        }


"""
Carbon -- data random interpolation library
"""
__VERSION__ = "1.0"
__DATE__ = "28/01/2023"

import numpy as _np
import pandas as _pd
from math import sqrt as _sqrt



class PathInterpolation():
    """
    interpolating path with additional data points respecting series volatility
    """
    __VERSION__ = __VERSION__
    __DATE__ = __DATE__

    @classmethod
    def gaussian_bridge(cls, a, b, sig, numsegs):
        """
        create a Gaussian bridge between two values
        
        :a:         the start value of the bridge
        :b:         the end value of the bridge
        :sig:       the lognormal volatility (0.1=10%); it is converted into
                    a standard deviation multiplying it with (a+b)/2
        :numsegs:   the number of segments to generate
        :returns:   numpy array of length numsegs+2, connecting the points a,b with
                    a path of numsegs segments, and a normal volatility of sig*(a+b)/2
        """
        if numsegs == 0:
            return _np.array([a, b])
        #stepsd = _sqrt(1/numsegs)
        increments  = _np.random.default_rng().normal(loc=0, scale=_sqrt(1/numsegs)*sig*(a+b)/2, size=numsegs)
        increments += (-sum(increments)+b-a)/numsegs
        increments  = _np.cumsum(_np.insert(increments, 0, a))
        assert abs(increments[-1]/b-1) < 1e-8, f"numerical divergence too large [{increments[-1]/b-1}]"
        increments[-1] = b
        return increments

    @classmethod
    def estimate_vol_np(cls, path, time_period=1):
        """
        estimates the volatility of a numpy path
        
        :path:          a numpy array of the path
        :time_period:   the time period covered by the path; default = 1
        :returns:       tuple normal volatility (stdev), lognormal volatility (sigma)
                        this is normalised that sd*dZ and sig*S*dZ where dZ is N(0,dt)
                        are suitable generating numbers for a path
        """
        steps = len(path)-1
        dt = time_period/steps
        sqrt_dt = _sqrt(dt)
        sd  = _np.std(_np.diff(path))/sqrt_dt
        sig = _np.std(_np.diff(_np.log(path)))/sqrt_dt
        return sd, sig

    @classmethod
    def estimate_vol(cls, path):
        """
        estimates the volatility of a pandas time series
        
        :path:          a pandas series whose index is a pandas timestamp
        :returns:       tuple normal volatility (stdev), lognormal volatility (sigma)
                        this is normalised that sd*dZ and sig*S*dZ where dZ is N(0,dt)
                        are suitable generating numbers for a path
        """
        time_period = (path.index[-1]-path.index[0])/_pd.Timedelta(days=1)/365.25
        return cls.estimate_vol_np(path, time_period)

    @classmethod
    def interpolate_segment(cls, path_segment, sig, period):
        """
        interpolates a path segment (exactly two points) with high frequency data
        
        :path_segment:    a pandas series of exactly two points, with timestamp index
        :sig:             the volatility sigma to use for the interpolation
        :period:          the desired length of the resampling period (as pandas Timedelta)
        :returns:         the resampled pandas series containing the first point and
                          the last point (must be removed for concatenation)
        """
        assert len(path_segment)==2, "path segment MUST be of length 2"
        path_td = path_segment.index[1]-path_segment.index[0]
        path_td_yrs = path_td/_pd.Timedelta(days=1)/365.25
        if period > path_td:
            print("[interpolate_segment] sample period too large; returning original")
            return path_segment.iloc[:1]
        numsegments = int(round(path_td/period,0))
        period_act = path_td/numsegments
        sample_path_np = cls.gaussian_bridge(path_segment[0], path_segment[1], sig*_sqrt(path_td_yrs), numsegments)
        sample_path_index = [path_segment.index[0] + i*period_act for i in range(numsegments+1)]
        return _pd.Series(sample_path_np, index=sample_path_index)

    @classmethod
    def interpolate(cls, path, period, sig=None, sigfctr=None, enable=True):
        """
        interpolates the path with high frequency data
        
        :path:         the path, as pandas series with pandas timestamps as index
        :obsperiod:    the desired resampling period, as pandas Timedelta (eg pd.Timedelta(minutes=1))
        :sig:          the sampling volatility (if None, estimated from the path)
        :sigfctr:      the factor by which the estimated vol is multiplied (default=1)
        :enable:       if False, simply returns `path` (allows use of this function in filter)
        :returns:      the interpolated path (or original, if enable=False)
        """
        if not enable:
            return path
        if sig is None:
            _, sig = cls.estimate_vol(path)
            if not sigfctr is None:
                sig *= sigfctr
            
        pathsegs     = (path.iloc[ix:ix+2] for ix in range(len(path)-1))
        interpolated = (cls.interpolate_segment(seg, sig, period)[:-1] for seg in pathsegs)
        newpath = _pd.concat([_pd.concat(interpolated), path.iloc[-1:]])
        return newpath 

    @staticmethod
    def Timedelta(*args, **kwargs):
        """alias for pd.Timedelta"""
        return _pd.Timedelta(*args, **kwargs)

    @classmethod
    def days(cls, days):
        """alias for Timedelta(days=days)"""
        return cls.Timedelta(days=days)

    @classmethod
    def hours(cls, hours):
        """alias for Timedelta(hours=hours)"""
        return cls.Timedelta(hours=hours)

    @classmethod
    def minutes(cls, minutes):
        """alias for Timedelta(minutes=minutes)"""
        return cls.Timedelta(minutes=minutes)

interpolate = PathInterpolation.interpolate



"""
represents a single, unidirectional carbon order and provides convenience methods for UI access

(c) Copyright Bprotocol foundation 2022. 
Licensed under MIT

VERSION HISTORY
- v1.3: order book helper functions (p_marg_f, yfromp_f, dyfromp_f, dyfromdx_f, dyfromdx_f, goalseek)
- v1.4: new methods: fromQxy, Q, Gamma, sellx, selly; yfromx_f, xfromy_f, p_eff_f, xint (1.3.1)
- v1.5: linked curves (beta); bidask, curves_by_pair_bidask, checks for B,S=0 (1.4.1)
- v1.6: linked curves incl trading (final), addliqy, tradeto; minor formula improvement (1.5.1)
- v1.7: integration with solidity testing (yzABS); bugfix (1.6.1)
- v1.8: from_SDK, bn2int, roundsd; some prettification
- v1.9: various additional properties representing curve parameters; used dataclass for yzABS; as_cpc
"""
__version__ = "1.9"
__date__ = "15/Mar/2023"

try:
    from .pair import CarbonPair
except:
    from pair import CarbonPair

from dataclasses import dataclass
from math import sqrt, floor, log10
from collections import namedtuple
from .sdk import Tokens
from .cpc import ConstantProductCurve as CPC

@dataclass
class CarbonOrderUI:
    """
    current state of a single Carbon order, for UI purposes
    
    properties that are parameters of the constructor
    :pair:    the corresponding token pair (specifically, its CarbonPair record)
    :tkn:     the token that this position is selling
    :B:       the B-parameter; B = sqrt pb_raw
    :S:       the S-parameter; S = sqrt pa_raw - Sqrt pb_raw
    :yint:    the y-intercept of the curve (also its current maximum capacity)
    :y:       the current y-coordinate on the curve (also current token holdings)
    :id:      id of this curve*
    :linked:  the linked curve object*

    *Beta; interace may still change in respect to ix, lix and linked as well 
    as the associated functions
    
    other properties set by the constructor
    :pa_raw:   the pa parameter in native quotation (dy/dx)
    :pb_raw:   ditto pb
    :pa:       the pa paramter in the quotation appropriate for the pair
    :pb:       ditto pb
    :pmin:     the min of pa, pb (in the quotation appropriate for the pair)
    :pmax:     the max of pa, pb (in the quotation appropriate for the pair)
    :reverseq: True if the pair is quoted in reverse order (ie pa_raw = 1/pa etc)
    
    NOTES
    - natively, prices are quoted in the convention dy/dx, where tkn is the quote asset;
    tkn is also the asset being sold, so the numeraire is always the asset being sold
    - pa, pb are read at the intercepts "left to right", so pa=py is is the y intercept price,
    and pb=px is the x intercept price 
    - the properties pa_raw and pb_raw correspond to the native pa, pb; the properties pa, pb
    are quoted in the correct currency conventions
    - the properties p_start = py = pa and p_end = px = pb are aliases
    - the property A is an alias of S (we use A in the smart contracts)
    """
    __VERSION__ = __version__
    __DATE__    = __date__
    
    pair: CarbonPair
    tkn: str
    B: float
    S: float
    yint: float
    y: float
    id: any = None
    linked: any = None

    def __post_init__(self):
        self.pair = CarbonPair(self.pair)
        self.tkn = self.tkn.upper()
        if not self.pair.has_token(self.tkn):
            raise ValueError("token not part of pair", self.tkn, self.pair)
        self.pb_raw = self.B * self.B
        self.pa_raw = (self.S + self.B)**2
        if self.pa_raw == 0 and self.pb_raw == 0:
            self.pa_raw = None
            self.pa_raw = None
            self.disabled = True
        else:
            self.disabled = False
        self.reverseq = True if self.pair.has_basetoken(self.tkn) else False

        if not self.disabled:
            self.pa = 1./self.pa_raw if self.reverseq else self.pa_raw
            self.pb = 1./self.pb_raw if self.reverseq else self.pb_raw
            if self.pa < self.pb:
                self.pmin = self.pa
                self.pmax = self.pb
            else:
                self.pmin = self.pb
                self.pmax = self.pa  
        else:
            self.pmin = None
            self.pmax = None        
    
    def set_id(self, id):
        """sets curve index; raises if curve ID already set (returns self)"""
        if not self.id is None:
            raise ValueError("Curve ID has already been set", id, self.id)
        self.id = id
        return self
    
    @property
    def lid(self):
        """returns ID of linked curve, or None"""
        if self.linked is None:
            return None
        return self.linked.id
        
    def set_linked(self, linked):
        """
        sets linked object

        :linked:            the linked curve (CarbonOrderUI object)
        :returns:           self
        """
        if not self.linked is None:
            #msg = f"Linked object has already been set {linked.id} {self.linked.id}"
            #print ("[set_linked]", msg)
            #print ("[set_linked] DEPRECIATION WARNING -- THIS WILL LEAD TO RAISE IN THE FUTURE")
            raise ValueError("Linked object has already been set", linked.id, self.linked.id)
            
        self.linked = linked
        # this code setting backlinks breaks previous assertions
        # if linked.linked is None:
        #     linked.linked = self
        return self

    @classmethod
    def from_BSy(cls, pair, tkn, B, S, yint, y):
        """
        alternative* constructor, taking B,S,y

        :pair:    the corresponding token pair (specifically, its CarbonPair record)
        :tkn:     the token that this order is selling
        :B:       the B-parameter; B = sqrt pb_raw
        :S:       the S-parameter; S = sqrt pa_raw - Sqrt pb_raw (also called A)
        :yint:    the y-intercept of the curve (also its current maximum capacity)
        :y:       the current y-coordinate on the curve (also current token holdings)

        *technically this is the same as the constructor, but use of `from_BSy` is recommended
        over use of the native class constructor in case of future implementation changes
        """
        return cls(
            pair=pair, 
            tkn=tkn, 
            B=B, 
            S=S, 
            yint=yint, 
            y=y
        )

    @classmethod
    def from_prices(cls, pair, tkn, pa, pb, yint, y, id=None):
        """
        alternative constructor, taking prices pa, pb and curve capacity yint
        
        :pair:    the corresponding token pair (specifically, its CarbonPair record)
        :tkn:     the token that this order is selling
        :pa:      the price at the y intercept, in quotation corresponding to the pair*
        :pb:      the price at the x intercept, in quotation corresponding to the pair*
        :yint:    the y-intercept of the curve (also its current maximum capacity)
        :y:       the current y-coordinate on the curve (also current token holdings)
        :id:      the curve ID (optional)
        
        *in their native quotation, pa, pb = -dy/dx at the y-intercept and x-intercept
        respectively; as the function y(x) is convex we must have pa >= pb; as this can
        be confusing in reverse quotation, so we correct by exchanging pa, pb if pb < pa
        """
        pair = CarbonPair(pair)
        if yint<0:
            raise ValueError(f"yint must be non-negative (yint={yint})", pair, tkn, pa, pb, yint, y)
        if y>yint:
            raise ValueError(f"y must not be bigger than yint (y={y}, yint={yint})", pair, tkn, pa, pb, yint, y)
        if y<0:
            raise ValueError(f"y must be non-negative (y={y})", pair, tkn, pa, pb, yint, y)

        if pair.has_basetoken(tkn):
            pa = 1./pa
            pb = 1./pb
        
        if pa < pb:
            print("[from_prices] exchanging pa, pb")
            paa = pa
            pa = pb
            pb = paa
            
        B = sqrt(pb)
        S = sqrt(pa) - sqrt(pb)
        return cls(
            pair=pair, 
            tkn=tkn, 
            B=B, 
            S=S, 
            yint=yint, 
            y=y,
            id=id,
        )
    
    @classmethod
    def from_Qxy(cls, pair, tkn, Q, xint, yint, y):
        """
        alternative constructor, taking convexity Q and curve capacities xint, yint
        """
        pair = CarbonPair(pair)
        if not pair.has_token(tkn):
            raise ValueError("token not part of the pair", tkn, pair)
        if xint<=0:
            raise ValueError("xint must be positive", xint)
        if yint<=0:
            raise ValueError("yint must be positive", yint)
        if y>yint:
            raise ValueError("y must not be bigger than yint (y={y}, yint={yint})", yint, y)
        if y<0:
            raise ValueError("y must be non-negative", y)

        p0 = yint/xint
        B = sqrt(p0*Q)
        S = sqrt(p0/Q) - B

        return cls(
            pair=pair, 
            tkn=tkn, 
            B=B, 
            S=S, 
            yint=yint, 
            y=y
        )

    @classmethod
    def from_order(cls, order):
        """
        alternative constructor, from an Order object

        :order:     the Order or CarbonOrderUI object
        """
        try: 
            yint = order.y_int
        except:
            yint = order.yint

        return cls(
            pair=order.pair, 
            tkn=order.tkn, 
            B=float(order.B), 
            S= float(order.S), 
            yint=float(yint), 
            y=float(order.y),
            id=order.id,
        )

    @classmethod 
    def _from_SDK0(cls, encodedorder, pair, tkn, dec, sx=None):
        """
        NOTAPI - alternative constructor, from a single order object coming from the SDK

        :encodedorder:    encoded order (AByz) object returned by the SDK, eg calling `getUserStrategies`
        :pair:            corresponding token pair (specifically, its CarbonPair record)    
        :tkn:             token that this order is selling
        :dec:             number of decimals to use for tkn
        :sx:              the scaling exponent to use (default: 48)
        :returns:         corresponding CarbonOrderUI object
        """
        if not encodedorder.get("encoded") is None:
            encodedorder = encodedorder["encoded"]
        AByz = cls.bn2intd(encodedorder)
        if sx is None: sx=48
        scalefctr = 2**sx
        decimalfctr = 10**dec
        S = AByz['A']/scalefctr  # S = A
        B = AByz['B']/scalefctr
        y = AByz['y']/decimalfctr
        z = AByz['z']/decimalfctr
        return cls.from_BSy(pair=pair, tkn=tkn, S=S, B=B, y=y, z=z)
    
    @classmethod
    def from_SDK(cls, sdkStrategy, nsd=None):
        """
        double constructor, returning two strategies from the SDK

        :sdkStrategy:   SDK strategy dict, eg returned by `getUserStrategies`
        :nsd:           number of significant decimals to round to (None=no rounding)
        :returns:       a tuple of two linked CarbonOrderUI objects
        """
        s = {k:cls.roundsd(float(sdkStrategy[k]), nsd) 
            for k in ['buyPriceLow', 'buyPriceHigh', 'buyBudget', 'sellPriceLow', 'sellPriceHigh', 'sellBudget']}
        sid = cls.bn2int(sdkStrategy["id"])
        if not {sdkStrategy["baseToken"], sdkStrategy["quoteToken"]} == {sdkStrategy["encoded"]["token0"], sdkStrategy["encoded"]["token1"]}:
            raise ValueError("tokens do not match encoded tokens", sdkStrategy)
        if sdkStrategy["baseToken"] == sdkStrategy["encoded"]["token0"]:
            tknbe = sdkStrategy["encoded"]["order0"]
            tknqe = sdkStrategy["encoded"]["order1"]
        else:
            tknbe = sdkStrategy["encoded"]["order1"]
            tknqe = sdkStrategy["encoded"]["order0"]
        tknqa = sdkStrategy["quoteToken"]
        tknqo = Tokens.byticker(tknqa, raiseonerror=False, returnnultkn=False)
        tknq = tknqo.T if not tknqo is None else "TKNQ"
        tknql = cls.bn2int(tknqe["y"]) / cls.bn2int(tknqe["z"])
        tknba = sdkStrategy["baseToken"]
        tknbo = Tokens.byticker(tknba, raiseonerror=False, returnnultkn=False)
        tknb = tknbo.T if not tknbo is None else "TKNB"
        tknbl = cls.bn2int(tknbe["y"]) / cls.bn2int(tknbe["z"])
        
        pair = CarbonPair(tknq=tknq, tknb=tknb)
        # the tkn in the constructor is the one that is being sold
        o_selltknb = cls.from_prices(pair=pair, tkn=tknb, pa=s["sellPriceLow"], 
            pb=s["sellPriceHigh"], yint=s["sellBudget"], y=s["sellBudget"]*tknbl, id=f"{sid}-s")
        o_buytknb  = cls.from_prices(pair=pair, tkn=tknq, pa=s["buyPriceHigh"], 
            pb=s["buyPriceLow"], yint=s["buyBudget"], y=s["buyBudget"]*tknql, id=f"{sid}-b")
        o_buytknb.set_linked(o_selltknb)
        o_selltknb.set_linked(o_buytknb)
        return o_buytknb, o_selltknb

    @property 
    def tkny(self):
        """the token on the y-axis (alias for tkn)"""
        return self.tkn
    
    @property 
    def tknx(self):
        """the token on the x-axis"""
        return self.pair.other(self.tkn)

    @property
    def bidask(self):
        """
        DONOTUSE -- returns BID or ASK depending on what type curve it is
        
        NOTE: BID sells base token, ASK sells quote token; this seems to be the wrong
        way round; however, if we change this we probably need to update the simulator
        sheet so we need to be careful -- DONOTUS
        """
        if self.tkny == self.pair.basetoken:
            return "BID"
        else:
            return "ASK"
    
    @property
    def p0(self):
        """
        the average or effective price of the range, p0 = sqrt(pa, pb)
        """
        return sqrt(self.pa*self.pb)
    
    @property
    def px(self):
        """alias for pb"""
        return self.pb 
    p_end = px

    @property
    def py(self):
        """alias for pa"""
        return self.pa
    p_start = py

    @property
    def A(self):
        """alias for S (notation used in smart contracts)"""
        return self.S
    
    @property
    def z(self):
        """alias for y (notation used in smart contracts)"""
        return self.y 

    @property
    def Q(self):
        """Q parameter = sqrt(Px/Py) < 1"""
        return sqrt(self.pb_raw/self.pa_raw)

    @property
    def Gamma(self):
        """Gamma parameter (also known as n) = 1 - sqrt(Q)"""
        return 1 - sqrt(self.Q)

    @staticmethod
    def Q_from_Gamma(Gamma):
        """Q(Gamma) = (1 - Gamma)^2"""
        return (1 - Gamma)**2
    
    @staticmethod
    def Gamma_from_Q(Q):
        """Gamma(Q) = 1 - sqrt(Q)"""
        return 1 - sqrt(Q)
    
    @staticmethod
    def asym_over_0(Gamma):
        """xasym/x0 or yasym/y0 = 1-1/Gamma < 0"""
        return 1 - 1/Gamma
    
    @staticmethod
    def int_over_0(Gamma):
        """xint/x0 or yint/y0 = (2-Gamma)/(1-Gamma) > 0"""
        return (2-Gamma)/(1-Gamma)
    
    @property
    def x0(self):
        """
        x0 alternative parameter (x0 y0 = p0)
        
        x0 = xint * (1-Gamma)/(2-Gamma)
        """
        return self.xint * (1-self.Gamma)/(2-self.Gamma)
    
    @property
    def y0(self):
        """
        alternative parameter (x0 y0 = p0)
        
        y0 = xint * (1-Gamma)/(2-Gamma)
        """
        return self.yint * (1-self.Gamma)/(2-self.Gamma)
    
    @property
    def p0_raw(self):
        """
        alternative parameter (x0 y0 = p0)
        """
        return self.y0 / self.x0
    
    @property
    def xasym(self):
        """
        alternative parameter (x asymptote); it is < 0
        """
        return self.x0 * (1-1/self.Gamma)

    @property
    def yasym(self):
        """
        alternative parameter (y asymptote); it is < 0
        """
        return self.y0 * (1-1/self.Gamma)
    
    @property
    def kappa(self):
        """
        alternative parameter x0 y0 / Gamma^2; see also kappa_bar

        Note: the invariant equation can be written as

        (x-xasym)*(y-yasym) = kappa
        """
        return self.x0*self.y0/(self.Gamma**2)
    
    @property
    def kappa_bar(self):
        """
        alternative parameter sqrt(x0 y0) / Gamma

        Note: the properly scaling invariant equation can be written as

        sqrt((x-xasym)*(y-yasym)) = kappa_bar
        """
        return sqrt(self.x0*self.y0)/self.Gamma
    
    @property
    def leverage_fctr(self):
        """
        leverage_fctr = 1/Gamma (see definition of kappa_bar)
        """
        return 1/self.Gamma
    
    @property
    def widthr(self):
        """the width ratio of the range, widthr = pmax/pmin"""
        return self.pmax/self.pmin
    
    @property
    def widthpc(self):
        """the percentage width of the range, widthpc = (pmax-pmin)/p0"""
        return (self.pmax-self.pmin)/self.p0

    @property
    def price_convention(self):
        """
        the price convention of the prices quoted
        """
        return self.pair.price_convention

    @property
    def price_convention_raw(self):
        """
        the price convention of pa_raw, pb_raw (ie dy/dx)
        """
        return f"{self.tkny} per {self.tknx}"

    def descr(self, full=False):
        """provides a description of the order and curve"""
        s1 = f"Sell {self.tkn} buy {self.pair.other(self.tkn)}"
        s2 = f"from {self.pa:.4f} to {self.pb:.4f} {self.price_convention}"
        s2 = f"from {self.pa:.4f} to {self.pb:.4f} {self.price_convention}"
        if full:
            s3 = f" ({self.y} {self.tkn} on curve, {self.y/self.yint*100:.0f}% of capacity)"
        else:
            s3 = ""
        return f"{s1} {s2}{s3}"

    @property
    def p_marg(self):
        """
        the current marginal price of the range (alias for p_margf)
        """
        return self.p_marg_f(0)

    def p_marg_f(self, dy=0, fullcurve=False, checkbounds=True, raiseonerror=False):
        """
        the marginal price function of the range

        :dy:            the negative(!) change in liquidity value, expressed as 
                        positive number; y* = y - dy
        :fullcurve:     if True, ignore curve state y and use yint, ie y* = yint - dy
        :checkbounds:   if True (default), check that y is in the right range
        :raiseonerror:  if True, raises on all errors; otherwise (default) may return None
        :returns:       marginal price a y*, p(y*) in the price convention of the pair
        """
        #dydx = ((self.B * self.yint + self.S * self.y) / self.yint)**2
        if self.disabled:
            return None

        y = self.yint - dy if fullcurve else self.y - dy
        #print(f"[p_marg_f] dy={dy} y={y} s.y={self.y} s.yint={self.yint}", fullcurve)
        if checkbounds:
            if dy < 0:
                if raiseonerror:
                    raise ValueError("Trade size dy must be a non-negative number", dy)
                return None
            if y < 0:
                if raiseonerror:
                    raise ValueError("Trade size dy too big, results in y<0", dy, y, self.y)
                return None
        
        if self.yint == 0:
            if not self.y == 0 and dy == 0:
                raise ValueError("If yint=0 you must also have y,dy=0", yint, y, dy)
            dydx = ((self.B + self.S))**2
        else:
            dydx = ((self.B + self.S * y/self.yint))**2
        result = dydx if self.pair.has_quotetoken(self.tkn) else 1/dydx
        return result
    
    def p_eff_f(self, dy=0, fullcurve=False, checkbounds=True, raiseonerror=False):
        """
        the effective price function of the range

        :dy:            the negative(!) change in liquidity value, expressed as 
                        positive number; y* = y - dy
        :fullcurve:     if True, ignore curve state y and use yint, ie y* = yint - dy
        :checkbounds:   if True (default), check that y is in the right range
        :raiseonerror:  if True, raises on all errors; otherwise (default) may return None
        :returns:       effective price between y (or yint) and y* in the price convention of the pair
       """
        #dydx = ((self.B * self.yint + self.S * self.y) / self.yint)**2
        if self.disabled:
            return None

        y0 = self.yint if fullcurve else self.y
        y = y0 - dy
        if checkbounds:
            if dy < 0:
                if raiseonerror:
                    raise ValueError("Trade size dy must be a non-negative number", dy)
                return None
            if y < 0:
                if raiseonerror:
                    raise ValueError("Trade size dy too big, results in y<0", dy, y, self.y)
                return None

        if self.yint == 0:
            if not y0 == 0 and dy == 0:
                raise ValueError("If yint=0 you must also have y,dy=0", yint, y, dy)
            dydx = ((self.B + self.S))**2
        else:
            if dy==0:
                return self.p_marg_f(self.yint-y, checkbounds=False, fullcurve=True)
            dydx = dy / (self.xfromy_f(y) - self.xfromy_f(y0))
        
        result = dydx if self.pair.has_quotetoken(self.tkn) else 1/dydx
        return result
    
    class PriceOutOfBoundsError(ValueError): pass
    class PriceOutOfBoundsErrorBeyondStart(PriceOutOfBoundsError): pass
    class PriceOutOfBoundsErrorBeyondEnd(PriceOutOfBoundsError): pass
    class PriceOutOfBoundsErrorBeyondMarg(PriceOutOfBoundsError): pass

    def yfromp_f(self, p, checkbounds=True, raiseonerror=False):
        """
        returns y as a function of the target marginal price

        :p:             the target marginal price, in the price convention of the pair
        :checkbounds:   if True (default), check that y is in the right range
        :raiseonerror:  if True, raises upon error, else return None
        :returns:       the y value at which the marginal price is achieved
                        if beyond the end it returns 0, if beyond start or current y None
        """
        # dydx = ((B * yint + S * y) / yint)**2 = (B + S y/yint)**2
        # y = yint * (sqrt(dydx) - B) / S
        dydx = p if self.pair.has_quotetoken(self.tkn) else 1/p
        #print(f"[yfromp_f] pa={self.pa_raw} dydx={dydx} pb={self.pb_raw}")

        if checkbounds:
            if dydx > self.pa_raw:
                if raiseonerror:
                    raise self.PriceOutOfBoundsErrorBeyondStart("Price out of bounds (beyond start)", p, self.pa)
                return None
            elif dydx < self.pb_raw:
                if raiseonerror:
                    raise self.PriceOutOfBoundsErrorBeyondEnd("Price out of bounds (beyond end)", p, self.pb)
                return 0
        y = self.yint * (sqrt(dydx) - self.B) / self.S
        if checkbounds:
            if y > self.y:
                if raiseonerror:
                    raise self.PriceOutOfBoundsErrorBeyondMarg("Price out of bounds (beyond marginal), hence target y > y", y, self.y )
                return None
        return y

    @property
    def xint(self):
        """
        x intercept, ie max x liquidity, ie x at y=0
        """
        #return self.yint**2 / (self.B**2*self.yint + self.B*self.S*self.yint)
        return self.yint / (self.B**2 + self.B*self.S)
        
    def dyfromp_f(self, p, checkbounds=True, raiseonerror=False):
        """
        returns dy = y_target - y as a function of the target marginal price

        :p:             the target marginal price, in the price convention of the pair
        :checkbounds:   if True (default), check that y is in the right range
        :raiseonerror:  if True, raises upon error, else return None
        :returns:       the (positive!) dy value at which the marginal price is achieved
                        in cases where yfromp_f returns none, this func returns 0
        """
        if self.B == 0 and self.S == 0:
            if raiseonerror:
                raise ValueError("Can't determine trade prices from an empty curve", self.B, self.S)
            return 0
        y = self.yfromp_f(p, checkbounds, raiseonerror)
        if y is None: return 0
        return self.y-y

    def dyfromdx_f(self, dx, checkbounds=True, raiseonerror=False):
        """
        calculates the amount dy SOLD by the AMM to RECEIVE an amount dx

        :dx:            the amount of x the AMM RECEIVES (a POSITIVE number*)
        :checkbounds:   if True (default), check that dy is in the right range
        :raiseonerror:  if True, raises upon error, else return None
        :returns:       the amount dy of y the AMM SELLS (a POSITIVE number*)

        *when checkbounds is False then we can have dx<0, corresponding to the 
        AMM SELLing x. In this case it returns a negative number, corresponding
        to the AMM BUYing y.
        """
        if self.B == 0 and self.S == 0:
            if raiseonerror:
                raise ValueError("Can't trade on an empty curve", self.B, self.S)
            return 0

        if checkbounds:
            if dx < 0:
                if raiseonerror:
                    raise ValueError("AMM buy amount dx must be a non-negative number", dx)
                return None
            # elif dx > self.y:
            #     if raiseonerror:
            #         raise ValueError("AMM sell amount dx must be within available liquidity", dx, self.y)
            #     return None
        
        num   =               (self.S*self.y + self.B*self.yint)**2
        #         -------------------------------------------------------------
        denom =   self.S*dx * (self.S*self.y + self.B*self.yint) + self.yint**2

        if checkbounds:
            if num < 0:
                if raiseonerror:
                    raise ValueError("AMM does not have enough y liquidity to purchase dx", y, dx, num)
                return None

        return dx * (num/denom) 

    def dxfromdy_f(self, dy, checkbounds=True, raiseonerror=False):
        """
        calculates the amount dx RECEIVED for a trade of dy

        :dy:            the amount of y the AMM sells (a POSITIVE number*)
        :checkbounds:   if True (default), check that dy is in the right range
        :raiseonerror:  if True, raises upon error, else return None
        :returns:       the amount of x the AMM receives (a POSITIVE number*)

        *when checkbounds is False then we can have dy<0, corresponding to the 
        AMM BUYing y. In this case it returns a negative number, corresponding
        to the AMM SELLing x.
        """
        if self.B == 0 and self.S == 0:
            if raiseonerror:
                raise ValueError("Can't trade on an empty curve", self.B, self.S)
            return 0

        if checkbounds:
            if dy < 0:
                if raiseonerror:
                    raise ValueError("AMM sell amount dy must be a non-negative number", dy)
                return None
            elif dy > self.y:
                if raiseonerror:
                    raise ValueError("AMM sell amount dy must be within available liquidity", dy, self.y)
                return None
        
        num   =                                   self.yint**2
        #       ----------------------------------------------------------------------------------
        denom = (self.S*self.y+self.B*self.yint) * (self.S*self.y+self.B*self.yint-self.S*dy)

        return dy*(num/denom)     
    
    def xfromy_f(self, y, checkbounds=True, raiseonerror=False):
        """
        the invariant function, expressed as x=f(y)

        :y:             the amount of y the AMM currently holds
        :checkbounds:   if True (default), check that dy is in the right range
        :raiseonerror:  if True, raises upon error, else return None
        :returns:       the corresponding x amount*

        *in Carbon, the funds received on this curve are not kept on this curve but 
        transferred to a linked curve; the linked curve may hold different amounts
        of x, eg because it has been seeded with x>0. In other words -- it does not
        make sense to look at absolute values of x, only of differences.    
        """
        if checkbounds:
            if y > self.yint:
                if raiseonerror:
                    raise ValueError("The value of y is out of bounds (y>yint)", y, self.yint)
                return None
            elif y < 0:
                if raiseonerror:
                    raise ValueError("The value of y is out of bounds (y<0)", y)
                return None

        num =                          self.yint * (self.yint - y)
        #        -----------------------------------------------------------------------------
        denom =  self.B**2*self.yint + self.B*self.S*y + self.B*self.S*self.yint + self.S**2*y
        return num/denom

    @property
    def x(self):
        """
        the current x value of the curve (=xfromy_f(self.y))
        """
        return self.xfromy_f(self.y)
    
    def yfromx_f(self, x, checkbounds=True, raiseonerror=False):
        """
        the invariant function, expressed as y=f(x)

        :x:             the theoretical* amount of x the AMM currently holds
        :checkbounds:   if True (default), check that dy is in the right range
        :raiseonerror:  if True, raises upon error, else return None
        :returns:       the corresponding      

        *in Carbon, the funds received on this curve are not kept on this curve but 
        transferred to a linked curve; the linked curve may hold different amounts
        of x, eg because it has been seeded with x>0. In other words -- it does not
        make sense to look at absolute values of x, only of differences.
        """
        if checkbounds:
            if x < 0:
                if raiseonerror:
                    raise ValueError("The value of x is out of bounds (x<0)", x)
                return None
            # if x < self.xint:
            #     if raiseonerror:
            #         raise ValueError("The value of x is out of bounds (x>xint; y<0)", x, result)
            #     return None 

        num =    self.yint*(-self.B**2*x - self.B*self.S*x + self.yint)
        #        ------------------------------------------------------
        denom =        self.B*self.S*x + self.S**2*x + self.yint

        y = num/denom
        if checkbounds:
            if y < 0:
                if raiseonerror:
                    raise ValueError("The value of x is out of bounds (x>xint; y<0)", x, xint, y)
                return None 
        return y

    @property
    def total_liquidity(self):
        """
        returns the total liquidity of the position, and the token in which it is quoted

        :returns:       (liquidity, token)
        """
        return self.y, self.tkn
    
    def liquidity_approx(self, price1, price2, tkn=None, asperc=False):
        """
        returns the approximate liquidity between start and end, in tkn

        :price1/2:      the start and end price of the range (in any order; quoted in price convention of pair)
        :tkn:           the token in which the liquidity is quoted (if None: base token)
        :asperc:        if True, return percentage total liquidity rather than tkn number; default is False
        :returns:       the liquidity in [price1, price2], quoted in tkn (or percent)
        """

        # ensure that price1 <= price2
        if price1 > price2:
            pp = price2
            price2 = price1
            price1 = pp

        if tkn is None:
            tkn = self.pair.basetoken
            #print(f"[liquidity_approx] token set to {tkn}")
            
        # # now we need to go through all the different arrangements of the range
        # # we first compute the percentage coverage, which simply is the percentage
            
        # # price1>pmax -> completely above the range
        # if price1 > self.pmax:
        #     perc = 0.
        
        # # price2<pmin -> completely below the range
        # elif price2 < self.pmin:
        #     perc = 0.

        # # price1 <= pmin
        # elif price1 <= self.pmin:

        #     # price1 <= min and price2 >= max -> 100%
        #     if price2 >= self.pmax:
        #         perc = 1.
        #     else:
        #         perc = (price2 - self.pmin)/(self.pmax - self.pmin)

        # # price2 >= pmax and price1>pmin
        # elif price2 >= self.pmax:

        #     perc = (self.pmax - price1)/(self.pmax - self.pmin)

        # # price2 < pmax and price1 > pmin
        # else:

        #     perc = (price2 - price1)/(self.pmax - self.pmin)

        # self.pmax == self.pmin -> 100% in range, otherwise out
        if self.disabled:
            return 0
            
        if self.pmax == self.pmin:
            if price1 == price2:
                perc2 = 1. if price1 == self.pmax else 0
            else:
                perc2 = 1. if price1 <= self.pmax and price2 > self.pmax else 0
            price = self.pmax
            #print(f"perc2={perc2}, price={price}", price1, price2, self.pmin, price1==self.pmax, price2==self.pmax)

        # price1 > self.pmax or price2 < self.pmin -> completely above or below the range
        elif price1 >= self.pmax or price2 <= self.pmin:
            perc2 = 0

        # alternatively: restrict p1,p2 to the ranges and calculate the coverage
        else:
            if price1 < self.pmin: price1 = self.pmin
            if price2 > self.pmax: price2 = self.pmax
            perc2 = (price2 - price1)/(self.pmax - self.pmin)
            price = sqrt(price1*price2)
        
        if asperc:
            return perc2
        
        if perc2 == 0:
            return 0

        # liquidity
        liq0 = self.y
        liq0_tkn = self.tkn

        # convert price into the correct quantity
        liq = self.pair.convert(amtfrom=liq0, tknfrom=liq0_tkn, tknto=tkn, price=price)
        
        return liq*perc2
    
    @staticmethod
    def goalseek(func, a, b, eps=None):
        """
        helper method: solves for x, a<x<b, such that func(x) == 0
        
        :func:    a function f(x), eg lambda x: x-3
        :a:       the lower bound
        :b:       the upper bound
        :eps:     precision, ie b/a value where goal seek returns
        :returns: the x value found
        """
        if eps is None:
            eps = 0.0000001
        if not a<b:
            raise ValueError("Bracketing value a must be smaller than b", a, b)
        fa = func(a)
        fb = func(b)
        if not fa*fb<0:
            raise ValueError("Sign of f(a) must be opposite of sign of f(b)", fa, fb, a, b)
            
        while 1:
            m = 0.5*(a+b)
            fm = func(m)
            if fm * fa > 0:
                a = m
            else:
                b = m
            
            #print(f"m={m}, m={m}, b={b}")
            if b/a-1 < eps:
                return m

    EPSILON = 1e-10
    def selly(self, dy, execute=True, allowneg=True, expandcurve=False, raiseonerror=False):
        """
        executes a trade selling dy for dx (dy given)

        :dy:            the amount of y to sell (a positive number)
        :execute:       if False, only display results but do not update the object
        :allowneg:      if True, negative dy numbers (=buying) are allowed
        :expandcurve:   if True, purchasing y beyond yint [sic] expands the curve to yint = y
                        only meaningful with allowneg = True
        :raiseonerror:  if True, error lead to raising on exception
        :returns:       a dict containing extensive information about the tx
        """
        if dy < 0:
            if not allowneg:
                if dy < -self.EPSILON:
                    if raiseonerror:
                        raise ValueError(f"Negative dy is not allowed (allowneg={allowneg})", dy)
                    return None
                else:
                    dy = 0

        if abs(dy) < self.EPSILON:
            dy = 0
            execute = False

        yold = self.y
        pold = self.p_marg
        ynew = self.y - dy
        pnew = self.p_marg_f(dy, checkbounds=False)

        if ynew < -self.EPSILON:
                if raiseonerror:
                    raise ValueError(f"Traded beyond capacity (yold={yold}, ynew={ynew}, dy={dy})")
                return None
        
        elif ynew > self.yint:
            if not expandcurve:
                if raiseonerror:
                    raise ValueError(f"Curve needs expanding and expanding not allowed (yold={yold}, ynew={ynew}, dy={dy}, expandcurve={expandcurve})")
                return None
            else:
                yintold = self.yint
                yint = ynew
                if execute:
                    self.yint = ynew
                curve_expanded = True
        else:
            curve_expanded = False
            yint = self.yint
            if ynew<0: ynew = 0  # see if ynew < -self.EPSILON

        dx = self.dxfromdy_f(dy, checkbounds=False, raiseonerror=True)
        if execute:
            self.y = ynew

        if self.linked:
            result_linked = self.linked.addliqy(dx)
        else:
            result_linked = None

        result = {
            "action":       "bysource[selly]",
            "executed":     execute,
            "y_old":        yold,
            "y":            ynew,
            "dy":           dy,
            "yint_old":     yintold if curve_expanded else None,
            "y_int":        yint,
            "expanded":     curve_expanded,
            "x":            self.xfromy_f(ynew),
            "dx":           dx,
            "tkny":         self.tkny,
            "tknx":         self.tknx,
            "tx":           f"Sell {abs(dy)} {self.tkny} buy {self.tknx}" if dx>0 else f"Buy {abs(dy)} {self.tkny} sell {self.tknx}",
            "dx/dy":        dx/dy if dy != 0 else None,
            "dy/dx":        dy/dx if dx != 0 else None,
            "pmarg_old":    pold,
            "pmarg":        pnew,
            "linked":       result_linked,
            "p":            None,
        } 
        result["p"] = result["dx/dy"] if self.pair.has_basetoken(self.tkny) else result["dy/dx"]
        return result

    def buyx(self, dx, execute=True, allowneg=True, expandcurve=False, raiseonerror=False):
        """
        executes a trade dy for dx (dx given)

        :dx:            the amount of x to buy (a positive number)
        :execute:       if False, only display results but do not update the object
        :allowneg:      if True, negative dy numbers (=buying) are allowed
        :expandcurve:   if True, purchasing y beyond yint [sic] expands the curve to yint = y
                        only meaningful with allowneg = True
        :raiseonerror:  if True, error lead to raising on exception
        :returns:       a dict containing extensive information about the tx
        """
        dy = self.dyfromdx_f(dx, checkbounds=False, raiseonerror=True)
        result = self.selly(dy, execute, allowneg, expandcurve, raiseonerror)
        result["action"] = "bytarget[buyx]"
        return result

    def tradeto(self, p_marg, execute=True):
        """
        executes a trade dy for dx (target marginal price p_marg given)

        :p:             the target marginal price
        :execute:       if False, only display results, but do not update the object
        :returns:       a dict containing extensive information about the tx
        """
        if self.yint == 0:
            self.yint = 1e-50 # selly fails on a fully empty curve, even with dy=0
        try:
            dy = self.dyfromp_f(p_marg, checkbounds=True, raiseonerror=True)
        except (self.PriceOutOfBoundsErrorBeyondStart, 
                self.PriceOutOfBoundsErrorBeyondMarg):
            dy = 0
        except self.PriceOutOfBoundsErrorBeyondEnd:
            dy = self.dyfromp_f(self.p_end, checkbounds=False, raiseonerror=False)
        result = self.selly(dy, execute, allowneg=False, expandcurve=False, raiseonerror=True)
        if result is None:
            raise RuntimeError(f"Unexpected None value at {p_marg}", dy, self.yint, self.y, self.p_marg, p_marg, self)
        result["action"] = "byprice[tradeto]"
        return result
        
    def addliqy(self, dy, expandcurve=True):
        """
        adds liquidity to the curve (typically called on a linked curve)

        :dy:            the amount of liquidity to be added (must be positive)
        :expandcurve:   if True (default), expand yint=y if need be
        :returns:       a dict containing extensive information about the tx
        """
        if dy < 0:
            raise ValueError("Liquidity amount dy must not be < 0", dy)
        
        newy = self.y + dy
        result = {
            "y_old":        self.y,
            "y":            newy,
            "dy":           dy,
            "yint_old":     self.yint,
            "yint":         self.yint,
            "tkny":         self.tkn,
            "pmarg_old":    self.p_marg,
            "pmarg":        None,
            "expanded":     False,
        }
        if newy > self.yint:
            if not expandcurve:
                raise ValueError("Must expand curve and expandcurve=False", self.y, dy, newy, self.yint)
            self.yint = newy
            result["yint"] = newy
            result["expanded"] = True
        self.y = newy
        result["pmarg"] = self.p_marg
        return result
        
    @staticmethod
    def curves_by_pair_bidask(curves):
        """
        DONOTUSE - sorts curves by pair, and then by bid/ask

        :curves:        an iterable of CarbonOrderUI curves
        :returns:       dict {
                            "pair": {
                                "ALL": [curve, ...],
                                "BID": [curve, ...],
                                "ASK": [curve, ...]
                            }
                            ...
                        }
        
        NOTE: apparently bidask is the wrong way round, but this is the way how currently
        the simulator expects it so we need to be careful changing it
        """
        pairs = set(r.pair.slashpair for r in curves.values())
        result = {
            pair:{ 
                "ALL": [v for v in curves.values() if v.pair.slashpair == pair],
                # NOTE THAT ASK AND BID HERE ARE EXCHANGED, SO IN PRINCIPLE WE COULD JUST CORRECT 
                # THE BIDASK FUNCTION AND CHANGE THEM BACK HERE...
                "ASK": [v for v in curves.values() if v.pair.slashpair == pair and v.bidask=="BID"],
                "BID": [v for v in curves.values() if v.pair.slashpair == pair and v.bidask=="ASK"],
            }
            for pair in pairs
        }            
        return result
    
    @property
    def as_cpc(self):
        """
        returns an equivalent constant product / virtual liquidity curve

        :returns:       a ConstantProductCurve object
        """
        p = self.p_marg
        if self.pair.has_basetoken(self.tkn): # we could also use self.bidask but this is broken at the moment TODO
            # selling the base token [ask curve?], therefore p=p_min and range ends add p_max
            p_min = p
            p_max = self.p_end
            #print("[as_cpc: ask curve]")
        else:
            # buying the base token [bid curve?], therefore p=p_max and range ends at p_min
            p_max = p
            p_min = self.p_end
            #print("[as_cpc: bid curve]")
    
        #print(p, p_min, p_max)

        return CPC.from_pkpp(
            p=p,
            k=self.kappa,
            p_min=p_min,
            p_max=p_max,
            pair=self.pair.slashpair
        )

    def yzABS(self, sx=0, verbose=False):
        """
        returns the parameters y,z,A,B,S needed for the smart contract*

        :sx:        the scaling exponent, scaling factor = 2**sx
        :verbose:   if True, prints detailed description of the calculation for audit
        :returns:   dataclass of ints (y,z,A,B,S)
                        :y:     y * 10**decy
                        :z:     yint * 10**decy
                        :A:     sqrt(10**(dec-dec)) * S * A
                        :B:     sqrt(...) * S * B
                        :S:     2**sx    

        *Important note: the parameter A in the return values corresponds to 
        the value S (or A) in this object; the value S in the return values
        corresponds to the scaling factor (!) 2**sx; it has nothing to do
        with the value S in this object      
        """
        if not self.pair.has_decimals:
            raise ValueError("pair must have decimals", self.pair)

        tkny = self.tkny
        tknx = self.tknx
        scale = 2**sx
        dec  = self.pair.decimals
        decy = dec[tkny]
        decx = dec[tknx]
        y_wei = self.y*10**decy
        z_wei = self.yint*10**decy
        B_ns = self.B * 10 ** ( (decy-decx)/2)
        A_ns = self.S * 10 ** ( (decy-decx)/2)
        yzABS = yzABSdata(y_wei, z_wei, int(A_ns*scale), int(B_ns*scale), scale)

        if verbose:
            print(f"[yzABS] pair={self.pair}, y={tkny}({decy}), x={tknx}({decx})")
            print(f"[yzABS] scale = 2**{sx} = {scale}")
            print(f"[yzABS] y={self.y} -> y_wei={y_wei} [{tkny}-wei]")
            print(f"[yzABS] yint={self.yint} -> z_wei={z_wei} [{tkny}-wei]")
            print(f"[yzABS] pa_raw={self.pa_raw} {tkny} per {tknx} -> {int(self.pa_raw*10**(decy-decx))} {tkny}-wei per {tknx}-wei")
            print(f"[yzABS] pb_raw={self.pb_raw} {tkny} per {tknx} -> {int(self.pb_raw*10**(decy-decx))} {tkny}-wei per {tknx}-wei")
            print(f"[yzABS] a={self.S} -> {A_ns} * scale = {yzABS.A}")
            print(f"[yzABS] b={self.B} -> {B_ns} * scale = {yzABS.B}")
            print(f"[yzABS] yzABS = {yzABS}")
        
        return yzABS

    @staticmethod
    def bn2int(item):
        """
        converts BigNumber to int (if item is a BND; else returns item)

        :returns: int(item) if bn BigNumber dict, else item
        """
        if not isinstance(item, dict): return item
        if not item.get("type") == 'BigNumber': return item
        return int(item["hex"], 16) 
    
    @classmethod
    def bn2intd(cls, dct):
        """
        applies bn2int to all values in a dict

        :returns: dict with int(item) if bn BigNumber dict, else item
        """
        return {k:cls.bn2int(v) for k,v in dct.items()}

    @staticmethod
    def roundsd(x, nsd=None):
        """
        rounds the value to significant decimals (returns x unmodified if not a number or zero)
        
        :x:      the value to be rounded
        :nsd:    number of significant decimals (None: don't round)
        """
        if nsd is None: return x
        try:
            magnitude = floor(log10(x))
        except:
            return x
        if magnitude <= nsd:
            return round(x,nsd-magnitude-1)
        else:
            shift = 10**(magnitude-nsd)
            return shift*round(x/shift,0)
    
    def __repr__(self):
        s1 = f"pair={str(self.pair)}, tkn={self.tkn}, B={self.B}, S={self.S}, yint={self.yint}, y={self.y}, id={self.id}"
        s2 = f"linked=<{self.linked.id}>" if self.linked else "linked=None"
        return f"{self.__class__.__name__}({s1}, {s2})"

@dataclass
class yzABSdata():
    """
    dataclass for yzABS data
    """
    y: int
    z: int
    A: int
    B: int
    S: int

    @property
    def astuple(self):
        return (self.y, self.z, self.A, self.B, self.S)
    
    @property
    def asdict(self):
        return dict(y=self.y, z=self.z, A=self.A, B=self.B, S=self.S)
